"""OpenTelemetry Anthropic instrumentation"""

import logging
import time
from typing import Callable, Collection, Optional

from opentelemetry import context as context_api
from opentelemetry._events import EventLogger, get_event_logger
from .config import Config
from .event_emitter import (
    emit_input_events,
    emit_response_events,
)
from .span_utils import (
    aset_input_attributes,
    aset_response_attributes,
    set_response_attributes,
)
from .streaming import (
    abuild_from_streaming_response,
    build_from_streaming_response,
)
from .utils import (
    acount_prompt_tokens_from_request,
    ashared_metrics_attributes,
    count_prompt_tokens_from_request,
    dont_throw,
    error_metrics_attributes,
    run_async,
    set_span_attribute,
    shared_metrics_attributes,
    should_emit_events,
)
from .streaming import (
    WrappedAsyncMessageStreamManager,
    WrappedMessageStreamManager,
)
from .version import __version__

from lmnr.opentelemetry_lib.tracing.context import get_current_context
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.instrumentation.utils import _SUPPRESS_INSTRUMENTATION_KEY, unwrap
from opentelemetry.metrics import Counter, Histogram, Meter, get_meter
from opentelemetry.semconv_ai import (
    SUPPRESS_LANGUAGE_MODEL_INSTRUMENTATION_KEY,
    LLMRequestTypeValues,
    Meters,
    SpanAttributes,
)
from opentelemetry.trace import Span, SpanKind, Tracer, get_tracer
from opentelemetry.trace.status import Status, StatusCode
from typing_extensions import Coroutine
from wrapt import wrap_function_wrapper

from anthropic._streaming import AsyncStream, Stream

logger = logging.getLogger(__name__)

_instruments = ("anthropic >= 0.3.11",)


WRAPPED_METHODS = [
    {
        "package": "anthropic.resources.completions",
        "object": "Completions",
        "method": "create",
        "span_name": "anthropic.completion",
    },
    {
        "package": "anthropic.resources.messages",
        "object": "Messages",
        "method": "create",
        "span_name": "anthropic.chat",
    },
    {
        "package": "anthropic.resources.messages",
        "object": "Messages",
        "method": "stream",
        "span_name": "anthropic.chat",
    },
    # This method is on an async resource, but is meant to be called as
    # an async context manager (async with), which we don't need to await;
    # thus, we wrap it with a sync wrapper
    {
        "package": "anthropic.resources.messages",
        "object": "AsyncMessages",
        "method": "stream",
        "span_name": "anthropic.chat",
    },
    # Beta API methods (regular Anthropic SDK)
    {
        "package": "anthropic.resources.beta.messages.messages",
        "object": "Messages",
        "method": "create",
        "span_name": "anthropic.chat",
    },
    {
        "package": "anthropic.resources.beta.messages.messages",
        "object": "Messages",
        "method": "stream",
        "span_name": "anthropic.chat",
    },
    # read note on async with above
    {
        "package": "anthropic.resources.beta.messages.messages",
        "object": "AsyncMessages",
        "method": "stream",
        "span_name": "anthropic.chat",
    },
    # Beta API methods (Bedrock SDK)
    {
        "package": "anthropic.lib.bedrock._beta_messages",
        "object": "Messages",
        "method": "create",
        "span_name": "anthropic.chat",
    },
    {
        "package": "anthropic.lib.bedrock._beta_messages",
        "object": "Messages",
        "method": "stream",
        "span_name": "anthropic.chat",
    },
    # read note on async with above
    {
        "package": "anthropic.lib.bedrock._beta_messages",
        "object": "AsyncMessages",
        "method": "stream",
        "span_name": "anthropic.chat",
    },
]

WRAPPED_AMETHODS = [
    {
        "package": "anthropic.resources.completions",
        "object": "AsyncCompletions",
        "method": "create",
        "span_name": "anthropic.completion",
    },
    {
        "package": "anthropic.resources.messages",
        "object": "AsyncMessages",
        "method": "create",
        "span_name": "anthropic.chat",
    },
    # Beta API async methods (regular Anthropic SDK)
    {
        "package": "anthropic.resources.beta.messages.messages",
        "object": "AsyncMessages",
        "method": "create",
        "span_name": "anthropic.chat",
    },
    # Beta API async methods (Bedrock SDK)
    {
        "package": "anthropic.lib.bedrock._beta_messages",
        "object": "AsyncMessages",
        "method": "create",
        "span_name": "anthropic.chat",
    },
]


def is_streaming_response(response):
    return isinstance(response, Stream) or isinstance(response, AsyncStream)


def is_stream_manager(response):
    """Check if response is a MessageStreamManager or AsyncMessageStreamManager"""
    try:
        from anthropic.lib.streaming._messages import (
            MessageStreamManager,
            AsyncMessageStreamManager,
        )

        return isinstance(response, (MessageStreamManager, AsyncMessageStreamManager))
    except ImportError:
        # Check by class name as fallback
        return (
            response.__class__.__name__ == "MessageStreamManager"
            or response.__class__.__name__ == "AsyncMessageStreamManager"
        )


@dont_throw
async def _aset_token_usage(
    span,
    anthropic,
    request,
    response,
    metric_attributes: dict = {},
    token_histogram: Histogram = None,
    choice_counter: Counter = None,
):
    # Handle with_raw_response wrapped responses first
    if response and hasattr(response, "parse") and callable(response.parse):
        try:
            response = response.parse()
        except Exception as e:
            logger.debug(f"Failed to parse with_raw_response: {e}")
            return

    usage = getattr(response, "usage", None) if response else None

    if usage:
        prompt_tokens = getattr(usage, "input_tokens", 0)
        cache_read_tokens = getattr(usage, "cache_read_input_tokens", 0) or 0
        cache_creation_tokens = getattr(usage, "cache_creation_input_tokens", 0) or 0
    else:
        prompt_tokens = await acount_prompt_tokens_from_request(anthropic, request)
        cache_read_tokens = 0
        cache_creation_tokens = 0

    input_tokens = prompt_tokens + cache_read_tokens + cache_creation_tokens

    if token_histogram and isinstance(input_tokens, int) and input_tokens >= 0:
        token_histogram.record(
            input_tokens,
            attributes={
                **metric_attributes,
                SpanAttributes.LLM_TOKEN_TYPE: "input",
            },
        )

    if usage:
        completion_tokens = getattr(usage, "output_tokens", 0)
    else:
        completion_tokens = 0
        if hasattr(anthropic, "count_tokens"):
            completion_attr = getattr(response, "completion", None)
            content_attr = getattr(response, "content", None)
            if completion_attr:
                completion_tokens = await anthropic.count_tokens(completion_attr)
            elif content_attr:
                completion_tokens = await anthropic.count_tokens(content_attr[0].text)

    if (
        token_histogram
        and isinstance(completion_tokens, int)
        and completion_tokens >= 0
    ):
        token_histogram.record(
            completion_tokens,
            attributes={
                **metric_attributes,
                SpanAttributes.LLM_TOKEN_TYPE: "output",
            },
        )

    total_tokens = input_tokens + completion_tokens

    choices = 0
    content_attr = getattr(response, "content", None)
    completion_attr = getattr(response, "completion", None)
    if isinstance(content_attr, list):
        choices = len(content_attr)
    elif completion_attr:
        choices = 1

    if choices > 0 and choice_counter:
        choice_counter.add(
            choices,
            attributes={
                **metric_attributes,
                SpanAttributes.LLM_RESPONSE_STOP_REASON: getattr(
                    response, "stop_reason", None
                ),
            },
        )

    set_span_attribute(span, SpanAttributes.LLM_USAGE_PROMPT_TOKENS, input_tokens)
    set_span_attribute(
        span, SpanAttributes.LLM_USAGE_COMPLETION_TOKENS, completion_tokens
    )
    set_span_attribute(span, SpanAttributes.LLM_USAGE_TOTAL_TOKENS, total_tokens)

    set_span_attribute(
        span, SpanAttributes.LLM_USAGE_CACHE_READ_INPUT_TOKENS, cache_read_tokens
    )
    set_span_attribute(
        span,
        SpanAttributes.LLM_USAGE_CACHE_CREATION_INPUT_TOKENS,
        cache_creation_tokens,
    )


@dont_throw
def _set_token_usage(
    span,
    anthropic,
    request,
    response,
    metric_attributes: dict = {},
    token_histogram: Histogram = None,
    choice_counter: Counter = None,
):
    # Handle with_raw_response wrapped responses first
    if response and hasattr(response, "parse") and callable(response.parse):
        try:
            response = response.parse()
        except Exception as e:
            logger.debug(f"Failed to parse with_raw_response: {e}")
            return

    usage = getattr(response, "usage", None) if response else None

    if usage:
        prompt_tokens = getattr(usage, "input_tokens", 0)
        cache_read_tokens = getattr(usage, "cache_read_input_tokens", 0) or 0
        cache_creation_tokens = getattr(usage, "cache_creation_input_tokens", 0) or 0
    else:
        prompt_tokens = count_prompt_tokens_from_request(anthropic, request)
        cache_read_tokens = 0
        cache_creation_tokens = 0

    input_tokens = prompt_tokens + cache_read_tokens + cache_creation_tokens

    if token_histogram and isinstance(input_tokens, int) and input_tokens >= 0:
        token_histogram.record(
            input_tokens,
            attributes={
                **metric_attributes,
                SpanAttributes.LLM_TOKEN_TYPE: "input",
            },
        )

    if usage:
        completion_tokens = getattr(usage, "output_tokens", 0)
    else:
        completion_tokens = 0
        if hasattr(anthropic, "count_tokens"):
            completion_attr = getattr(response, "completion", None)
            content_attr = getattr(response, "content", None)
            if completion_attr:
                completion_tokens = anthropic.count_tokens(completion_attr)
            elif content_attr:
                completion_tokens = anthropic.count_tokens(content_attr[0].text)

    if (
        token_histogram
        and isinstance(completion_tokens, int)
        and completion_tokens >= 0
    ):
        token_histogram.record(
            completion_tokens,
            attributes={
                **metric_attributes,
                SpanAttributes.LLM_TOKEN_TYPE: "output",
            },
        )
    total_tokens = input_tokens + completion_tokens
    choices = 0

    content_attr = getattr(response, "content", None)
    completion_attr = getattr(response, "completion", None)
    if isinstance(content_attr, list):
        choices = len(content_attr)
    elif completion_attr:
        choices = 1
    if choices > 0 and choice_counter:
        choice_counter.add(
            choices,
            attributes={
                **metric_attributes,
                SpanAttributes.LLM_RESPONSE_STOP_REASON: getattr(
                    response, "stop_reason", None
                ),
            },
        )

    set_span_attribute(span, SpanAttributes.LLM_USAGE_PROMPT_TOKENS, input_tokens)
    set_span_attribute(
        span, SpanAttributes.LLM_USAGE_COMPLETION_TOKENS, completion_tokens
    )
    set_span_attribute(span, SpanAttributes.LLM_USAGE_TOTAL_TOKENS, total_tokens)

    set_span_attribute(
        span, SpanAttributes.LLM_USAGE_CACHE_READ_INPUT_TOKENS, cache_read_tokens
    )
    set_span_attribute(
        span,
        SpanAttributes.LLM_USAGE_CACHE_CREATION_INPUT_TOKENS,
        cache_creation_tokens,
    )


def _with_chat_telemetry_wrapper(func):
    """Helper for providing tracer for wrapper functions. Includes metric collectors."""

    def _with_chat_telemetry(
        tracer,
        token_histogram,
        choice_counter,
        duration_histogram,
        exception_counter,
        event_logger,
        to_wrap,
    ):
        def wrapper(wrapped, instance, args, kwargs):
            return func(
                tracer,
                token_histogram,
                choice_counter,
                duration_histogram,
                exception_counter,
                event_logger,
                to_wrap,
                wrapped,
                instance,
                args,
                kwargs,
            )

        return wrapper

    return _with_chat_telemetry


def _create_metrics(meter: Meter):
    token_histogram = meter.create_histogram(
        name=Meters.LLM_TOKEN_USAGE,
        unit="token",
        description="Measures number of input and output tokens used",
    )

    choice_counter = meter.create_counter(
        name=Meters.LLM_GENERATION_CHOICES,
        unit="choice",
        description="Number of choices returned by chat completions call",
    )

    duration_histogram = meter.create_histogram(
        name=Meters.LLM_OPERATION_DURATION,
        unit="s",
        description="GenAI operation duration",
    )

    exception_counter = meter.create_counter(
        name=Meters.LLM_ANTHROPIC_COMPLETION_EXCEPTIONS,
        unit="time",
        description="Number of exceptions occurred during chat completions",
    )

    return token_histogram, choice_counter, duration_histogram, exception_counter


@dont_throw
def _handle_input(span: Span, event_logger: Optional[EventLogger], kwargs):
    if should_emit_events() and event_logger:
        emit_input_events(event_logger, kwargs)
    else:
        if not span.is_recording():
            return
        run_async(aset_input_attributes(span, kwargs))


@dont_throw
async def _ahandle_input(span: Span, event_logger: Optional[EventLogger], kwargs):
    if should_emit_events() and event_logger:
        emit_input_events(event_logger, kwargs)
    else:
        if not span.is_recording():
            return
        await aset_input_attributes(span, kwargs)


@dont_throw
def _handle_response(span: Span, event_logger: Optional[EventLogger], response):
    if should_emit_events():
        emit_response_events(event_logger, response)
    else:
        if not span.is_recording():
            return
        set_response_attributes(span, response)


@dont_throw
async def _ahandle_response(span: Span, event_logger: Optional[EventLogger], response):
    if should_emit_events():
        emit_response_events(event_logger, response)
    else:
        if not span.is_recording():
            return

        await aset_response_attributes(span, response)


@_with_chat_telemetry_wrapper
def _wrap(
    tracer: Tracer,
    token_histogram: Histogram,
    choice_counter: Counter,
    duration_histogram: Histogram,
    exception_counter: Counter,
    event_logger: Optional[EventLogger],
    to_wrap,
    wrapped,
    instance,
    args,
    kwargs,
):
    """Instruments and calls every function defined in TO_WRAP."""
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY) or context_api.get_value(
        SUPPRESS_LANGUAGE_MODEL_INSTRUMENTATION_KEY
    ):
        return wrapped(*args, **kwargs)

    name = to_wrap.get("span_name")
    span = tracer.start_span(
        name,
        kind=SpanKind.CLIENT,
        attributes={
            SpanAttributes.LLM_SYSTEM: "anthropic",
            SpanAttributes.LLM_REQUEST_TYPE: LLMRequestTypeValues.COMPLETION.value,
        },
        context=get_current_context(),
    )

    _handle_input(span, event_logger, kwargs)

    start_time = time.time()
    try:
        response = wrapped(*args, **kwargs)
    except Exception as e:  # pylint: disable=broad-except
        end_time = time.time()
        attributes = error_metrics_attributes(e)

        if duration_histogram:
            duration = end_time - start_time
            duration_histogram.record(duration, attributes=attributes)

        if exception_counter:
            exception_counter.add(1, attributes=attributes)

        raise e

    end_time = time.time()

    if is_streaming_response(response):
        return build_from_streaming_response(
            span,
            response,
            instance._client,
            start_time,
            token_histogram,
            choice_counter,
            duration_histogram,
            exception_counter,
            event_logger,
            kwargs,
        )
    elif is_stream_manager(response):
        if response.__class__.__name__ == "AsyncMessageStreamManager":
            return WrappedAsyncMessageStreamManager(
                response,
                span,
                instance._client,
                start_time,
                token_histogram,
                choice_counter,
                duration_histogram,
                exception_counter,
                event_logger,
                kwargs,
            )
        else:
            return WrappedMessageStreamManager(
                response,
                span,
                instance._client,
                start_time,
                token_histogram,
                choice_counter,
                duration_histogram,
                exception_counter,
                event_logger,
                kwargs,
            )
    elif response:
        try:
            metric_attributes = shared_metrics_attributes(response)

            if duration_histogram:
                duration = time.time() - start_time
                duration_histogram.record(
                    duration,
                    attributes=metric_attributes,
                )

            _handle_response(span, event_logger, response)
            if span.is_recording():
                _set_token_usage(
                    span,
                    instance._client,
                    kwargs,
                    response,
                    metric_attributes,
                    token_histogram,
                    choice_counter,
                )
        except Exception as ex:  # pylint: disable=broad-except
            logger.warning(
                "Failed to set response attributes for anthropic span, error: %s",
                str(ex),
            )

        if span.is_recording():
            span.set_status(Status(StatusCode.OK))
    span.end()
    return response


@_with_chat_telemetry_wrapper
async def _awrap(
    tracer,
    token_histogram: Histogram,
    choice_counter: Counter,
    duration_histogram: Histogram,
    exception_counter: Counter,
    event_logger: Optional[EventLogger],
    to_wrap,
    wrapped,
    instance,
    args,
    kwargs,
):
    """Instruments and calls every function defined in TO_WRAP."""
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY) or context_api.get_value(
        SUPPRESS_LANGUAGE_MODEL_INSTRUMENTATION_KEY
    ):
        return await wrapped(*args, **kwargs)

    name = to_wrap.get("span_name")
    span = tracer.start_span(
        name,
        kind=SpanKind.CLIENT,
        attributes={
            SpanAttributes.LLM_SYSTEM: "anthropic",
            SpanAttributes.LLM_REQUEST_TYPE: LLMRequestTypeValues.COMPLETION.value,
        },
        context=get_current_context(),
    )
    await _ahandle_input(span, event_logger, kwargs)

    start_time = time.time()
    try:
        response = await wrapped(*args, **kwargs)
    except Exception as e:  # pylint: disable=broad-except
        end_time = time.time()
        attributes = error_metrics_attributes(e)

        if duration_histogram:
            duration = end_time - start_time
            duration_histogram.record(duration, attributes=attributes)

        if exception_counter:
            exception_counter.add(1, attributes=attributes)

        raise e

    if is_streaming_response(response):
        return abuild_from_streaming_response(
            span,
            response,
            instance._client,
            start_time,
            token_histogram,
            choice_counter,
            duration_histogram,
            exception_counter,
            event_logger,
            kwargs,
        )
    elif is_stream_manager(response):
        if response.__class__.__name__ == "AsyncMessageStreamManager":
            return WrappedAsyncMessageStreamManager(
                response,
                span,
                instance._client,
                start_time,
                token_histogram,
                choice_counter,
                duration_histogram,
                exception_counter,
                event_logger,
                kwargs,
            )
        else:
            return WrappedMessageStreamManager(
                response,
                span,
                instance._client,
                start_time,
                token_histogram,
                choice_counter,
                duration_histogram,
                exception_counter,
                event_logger,
                kwargs,
            )
    elif response:
        metric_attributes = await ashared_metrics_attributes(response)

        if duration_histogram:
            duration = time.time() - start_time
            duration_histogram.record(
                duration,
                attributes=metric_attributes,
            )

        await _ahandle_response(span, event_logger, response)

        if span.is_recording():
            await _aset_token_usage(
                span,
                instance._client,
                kwargs,
                response,
                metric_attributes,
                token_histogram,
                choice_counter,
            )
            span.set_status(Status(StatusCode.OK))
    span.end()
    return response


def is_metrics_enabled() -> bool:
    return False


class AnthropicInstrumentor(BaseInstrumentor):
    """An instrumentor for Anthropic's client library."""

    def __init__(
        self,
        enrich_token_usage: bool = False,
        exception_logger=None,
        use_legacy_attributes: bool = True,
        get_common_metrics_attributes: Callable[[], dict] = lambda: {},
        upload_base64_image: Optional[
            Callable[[str, str, str, str], Coroutine[None, None, str]]
        ] = None,
    ):
        super().__init__()
        Config.exception_logger = exception_logger
        Config.enrich_token_usage = enrich_token_usage
        Config.get_common_metrics_attributes = get_common_metrics_attributes
        Config.upload_base64_image = upload_base64_image
        Config.use_legacy_attributes = use_legacy_attributes

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        tracer_provider = kwargs.get("tracer_provider")
        tracer = get_tracer(__name__, __version__, tracer_provider)

        # meter and counters are inited here
        meter_provider = kwargs.get("meter_provider")
        meter = get_meter(__name__, __version__, meter_provider)

        if is_metrics_enabled():
            (
                token_histogram,
                choice_counter,
                duration_histogram,
                exception_counter,
            ) = _create_metrics(meter)
        else:
            (
                token_histogram,
                choice_counter,
                duration_histogram,
                exception_counter,
            ) = (None, None, None, None)

        # event_logger is inited here
        event_logger = None

        if not Config.use_legacy_attributes:
            event_logger_provider = kwargs.get("event_logger_provider")
            event_logger = get_event_logger(
                __name__, __version__, event_logger_provider=event_logger_provider
            )

        for wrapped_method in WRAPPED_METHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            wrap_method = wrapped_method.get("method")

            try:
                wrap_function_wrapper(
                    wrap_package,
                    f"{wrap_object}.{wrap_method}",
                    _wrap(
                        tracer,
                        token_histogram,
                        choice_counter,
                        duration_histogram,
                        exception_counter,
                        event_logger,
                        wrapped_method,
                    ),
                )
                logger.debug(
                    f"Successfully wrapped {wrap_package}.{wrap_object}.{wrap_method}"
                )
            except Exception as e:
                logger.debug(
                    f"Failed to wrap {wrap_package}.{wrap_object}.{wrap_method}: {e}"
                )
            except ModuleNotFoundError:
                pass  # that's ok, we don't want to fail if some methods do not exist

        for wrapped_method in WRAPPED_AMETHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            wrap_method = wrapped_method.get("method")
            try:
                wrap_function_wrapper(
                    wrap_package,
                    f"{wrap_object}.{wrap_method}",
                    _awrap(
                        tracer,
                        token_histogram,
                        choice_counter,
                        duration_histogram,
                        exception_counter,
                        event_logger,
                        wrapped_method,
                    ),
                )
            except Exception:
                pass  # that's ok, we don't want to fail if some methods do not exist

    def _uninstrument(self, **kwargs):
        for wrapped_method in WRAPPED_METHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            unwrap(
                f"{wrap_package}.{wrap_object}",
                wrapped_method.get("method"),
            )
        for wrapped_method in WRAPPED_AMETHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            unwrap(
                f"{wrap_package}.{wrap_object}",
                wrapped_method.get("method"),
            )



from lmnr.opentelemetry_lib.decorators import (
    observe_base,
    async_observe_base,
)

from typing import Any, Callable, Coroutine, Literal, TypeVar, overload
from typing_extensions import ParamSpec

from lmnr.sdk.log import get_default_logger
from lmnr.sdk.types import TraceType

from .utils import is_async
import os

logger = get_default_logger(__name__)

P = ParamSpec("P")
R = TypeVar("R")


# Overload for synchronous functions
@overload
def observe(
    *,
    name: str | None = None,
    session_id: str | None = None,
    user_id: str | None = None,
    ignore_input: bool = False,
    ignore_output: bool = False,
    span_type: Literal["DEFAULT", "LLM", "TOOL"] = "DEFAULT",
    ignore_inputs: list[str] | None = None,
    input_formatter: Callable[..., str] | None = None,
    output_formatter: Callable[..., str] | None = None,
    metadata: dict[str, Any] | None = None,
    tags: list[str] | None = None,
    preserve_global_context: bool = False,
    rollout_entrypoint: bool = False,
) -> Callable[[Callable[P, R]], Callable[P, R]]: ...


# Overload for asynchronous functions
@overload
def observe(
    *,
    name: str | None = None,
    session_id: str | None = None,
    user_id: str | None = None,
    ignore_input: bool = False,
    ignore_output: bool = False,
    span_type: Literal["DEFAULT", "LLM", "TOOL"] = "DEFAULT",
    ignore_inputs: list[str] | None = None,
    input_formatter: Callable[..., str] | None = None,
    output_formatter: Callable[..., str] | None = None,
    metadata: dict[str, Any] | None = None,
    tags: list[str] | None = None,
    preserve_global_context: bool = False,
    rollout_entrypoint: bool = False,
) -> Callable[
    [Callable[P, Coroutine[Any, Any, R]]], Callable[P, Coroutine[Any, Any, R]]
]: ...


# Implementation
def observe(
    *,
    name: str | None = None,
    session_id: str | None = None,
    user_id: str | None = None,
    ignore_input: bool = False,
    ignore_output: bool = False,
    span_type: Literal["DEFAULT", "LLM", "TOOL"] = "DEFAULT",
    ignore_inputs: list[str] | None = None,
    input_formatter: Callable[..., str] | None = None,
    output_formatter: Callable[..., str] | None = None,
    metadata: dict[str, Any] | None = None,
    tags: list[str] | None = None,
    preserve_global_context: bool = False,
    rollout_entrypoint: bool = False,
):
    # Return type is determined by overloads above
    """The main decorator entrypoint for Laminar. This is used to wrap
    functions and methods to create spans.

    Args:
        name (str | None, optional): Name of the span. Function name is used if\
            not specified. Defaults to None.
        session_id (str | None, optional): Session ID to associate with the\
            span and the following context. Defaults to None.
        user_id (str | None, optional): User ID to associate with the span and\
            the following context. This is different from ID of a Laminar user.
            Defaults to None.
        ignore_input (bool, optional): Whether to ignore ALL input of the\
            wrapped function. Defaults to False.
        ignore_output (bool, optional): Whether to ignore ALL output of the\
            wrapped function. Defaults to False.
        span_type (Literal["DEFAULT", "LLM", "TOOL"], optional): Type of the span.
            Defaults to "DEFAULT".
        ignore_inputs (list[str] | None, optional): List of input keys to\
            ignore. For example, if the wrapped function takes three arguments\
            def foo(a, b, `sensitive_data`), and you want to ignore the\
            `sensitive_data` argument, you can pass ["sensitive_data"] to\
            this argument. Defaults to None.
        input_formatter (Callable[P, str] | None, optional): A custom function\
            to format the input of the wrapped function. This function should\
            accept the same parameters as the wrapped function and return a string.\
            All function arguments are passed to this function. Ignored if\
            `ignore_input` is True. Does not respect `ignore_inputs` argument.
            Defaults to None.
        output_formatter (Callable[[R], str] | None, optional): A custom function\
            to format the output of the wrapped function. This function should\
            accept a single parameter (the return value of the wrapped function)\
            and return a string. Ignored if `ignore_output` is True.\
            Defaults to None.
        metadata (dict[str, Any] | None, optional): Metadata to associate with\
            the trace. Must be JSON serializable. Defaults to None.
        tags (list[str] | None, optional): Tags to associate with the trace.
            Defaults to None.
        preserve_global_context (bool, optional): Whether to preserve the global\
            OpenTelemetry context. If set to True, Laminar spans will continue\
            traces started in the global context. Defaults to False.
        rollout_entrypoint (bool, optional): Whether to mark this function as a\
            rollout entrypoint. When True and in rollout mode, the function will\
            be registered for rollout execution via the CLI. Defaults to False.
    Raises:
        Exception: re-raises the exception if the wrapped function raises an\
            exception

    Returns:
        R: Returns the result of the wrapped function
    """

    def decorator(
        func: Callable[P, R] | Callable[P, Coroutine[Any, Any, R]],
    ) -> Callable[P, R] | Callable[P, Coroutine[Any, Any, R]]:
        # Get rollout session ID from environment if in rollout mode
        rollout_session_id = os.environ.get("LMNR_ROLLOUT_SESSION_ID")

        association_properties = {}
        if session_id is not None:
            association_properties["session_id"] = session_id
        if user_id is not None:
            association_properties["user_id"] = user_id
        if rollout_session_id is not None:
            association_properties["rollout_session_id"] = rollout_session_id
        if span_type in ["EVALUATION", "EXECUTOR", "EVALUATOR"]:
            association_properties["trace_type"] = TraceType.EVALUATION.value
        if tags is not None:
            if not isinstance(tags, list) or not all(
                isinstance(tag, str) for tag in tags
            ):
                logger.warning("Tags must be a list of strings. Tags will be ignored.")
            else:
                # list(set(tags)) to deduplicate tags
                association_properties["tags"] = list(set(tags))
        if input_formatter is not None and ignore_input:
            logger.warning(
                f"observe, function {func.__name__}: Input formatter"
                " is ignored because `ignore_input` is True. Specify only one of"
                " `ignore_input` or `input_formatter`."
            )
        if input_formatter is not None and ignore_inputs is not None:
            logger.warning(
                f"observe, function {func.__name__}: Both input formatter and"
                " `ignore_inputs` are specified. Input formatter"
                " will pass all arguments to the formatter regardless of"
                " `ignore_inputs`."
            )
        if output_formatter is not None and ignore_output:
            logger.warning(
                f"observe, function {func.__name__}: Output formatter"
                " is ignored because `ignore_output` is True. Specify only one of"
                " `ignore_output` or `output_formatter`."
            )

        # Merge rollout.session_id into metadata if in rollout mode
        merged_metadata = metadata.copy() if metadata else {}
        if rollout_session_id is not None:
            merged_metadata["rollout.session_id"] = rollout_session_id

        if is_async(func):
            wrapped_func = async_observe_base(
                name=name,
                ignore_input=ignore_input,
                ignore_output=ignore_output,
                span_type=span_type,
                metadata=merged_metadata if merged_metadata else None,
                ignore_inputs=ignore_inputs,
                input_formatter=input_formatter,
                output_formatter=output_formatter,
                association_properties=association_properties,
                preserve_global_context=preserve_global_context,
            )(func)
        else:
            wrapped_func = observe_base(
                name=name,
                ignore_input=ignore_input,
                ignore_output=ignore_output,
                span_type=span_type,
                metadata=merged_metadata if merged_metadata else None,
                ignore_inputs=ignore_inputs,
                input_formatter=input_formatter,
                output_formatter=output_formatter,
                association_properties=association_properties,
                preserve_global_context=preserve_global_context,
            )(func)

        # Handle rollout entrypoint registration (after wrapping)
        if rollout_entrypoint:
            from lmnr.sdk.rollout_control import is_rollout_mode, register_entrypoint

            if is_rollout_mode():
                # Register the WRAPPED function for rollout execution
                entrypoint_name = name if name is not None else func.__name__
                register_entrypoint(entrypoint_name, wrapped_func)

        return wrapped_func

    return decorator



from typing import Callable, Optional

from opentelemetry._events import EventLogger


class Config:
    enrich_token_usage = False
    enrich_assistant = False
    exception_logger = None
    get_common_metrics_attributes: Callable[[], dict] = lambda: {}
    upload_base64_image: Callable[[str, str, str], str] = (
        lambda trace_id, span_id, base64_image_url: ""
    )
    enable_trace_context_propagation: bool = True
    use_legacy_attributes = True
    event_logger: Optional[EventLogger] = None



"""OpenTelemetry Langgraph instrumentation"""

import json
import logging
from typing import Collection

from .utils import (
    with_tracer_wrapper,
)

from langchain_core.runnables.graph import Graph
from opentelemetry.trace import Tracer
from wrapt import wrap_function_wrapper
from opentelemetry.trace import get_tracer
from opentelemetry.context import get_value, attach, set_value

from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.instrumentation.utils import unwrap


logger = logging.getLogger(__name__)

_instruments = ("langgraph >= 0.1.0",)


@with_tracer_wrapper
def wrap_pregel_stream(tracer: Tracer, to_wrap, wrapped, instance, args, kwargs):
    graph: Graph = instance.get_graph()
    nodes = [
        {
            "id": node.id,
            "name": node.name,
            "metadata": node.metadata,
        }
        for node in graph.nodes.values()
    ]
    edges = [
        {
            "source": edge.source,
            "target": edge.target,
            "conditional": edge.conditional,
        }
        for edge in graph.edges
    ]
    d = {
        "langgraph.edges": json.dumps(edges),
        "langgraph.nodes": json.dumps(nodes),
    }
    association_properties = get_value("lmnr.langgraph.graph") or {}
    association_properties.update(d)
    attach(set_value("lmnr.langgraph.graph", association_properties))
    return wrapped(*args, **kwargs)


@with_tracer_wrapper
async def async_wrap_pregel_stream(
    tracer: Tracer, to_wrap, wrapped, instance, args, kwargs
):
    graph: Graph = await instance.aget_graph()
    nodes = [
        {
            "id": node.id,
            "name": node.name,
            "metadata": node.metadata,
        }
        for node in graph.nodes.values()
    ]
    edges = [
        {
            "source": edge.source,
            "target": edge.target,
            "conditional": edge.conditional,
        }
        for edge in graph.edges
    ]

    d = {
        "langgraph.edges": json.dumps(edges),
        "langgraph.nodes": json.dumps(nodes),
    }
    association_properties = get_value("lmnr.langgraph.graph") or {}
    association_properties.update(d)
    attach(set_value("lmnr.langgraph.graph", association_properties))

    async for item in wrapped(*args, **kwargs):
        yield item


class LanggraphInstrumentor(BaseInstrumentor):
    """An instrumentor for Langgraph."""

    def __init__(self):
        super().__init__()

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        tracer_provider = kwargs.get("tracer_provider")
        tracer = get_tracer(__name__, "0.0.1a0", tracer_provider)

        wrap_function_wrapper(
            module="langgraph.pregel",
            name="Pregel.stream",
            wrapper=wrap_pregel_stream(tracer, "Pregel.stream"),
        )
        wrap_function_wrapper(
            module="langgraph.pregel",
            name="Pregel.astream",
            wrapper=async_wrap_pregel_stream(tracer, "Pregel.astream"),
        )

    def _uninstrument(self, **kwargs):
        unwrap(
            module="langgraph.pregel",
            name="Pregel.stream",
        )
        unwrap(
            module="langgraph.pregel",
            name="Pregel.astream",
        )



"""OpenTelemetry Daytona SDK instrumentation

This module instruments the Daytona SDK to capture traces and logs for
execute_session_command calls.

The instrumentation handles both synchronous and asynchronous commands:

Synchronous commands (run_async/var_async=False):
1. Create a span when execute_session_command is called
2. Execute the command and wait for completion
3. End the span after the command returns
4. Emit logs immediately from response.stdout/stderr

Asynchronous commands (run_async/var_async=True):
1. Create a span when execute_session_command is called
2. Execute the command (returns immediately with cmd_id)
3. End the span after execute_session_command returns
4. Start background log streaming to capture stdout/stderr as they arrive
5. Emit OpenTelemetry logs for each line using the Logs API

Note: The module path for the Daytona SDK may need adjustment based on the
actual package structure. Update WRAPPED_METHODS if the module path differs.
"""

from typing import Collection
from importlib.metadata import version

from lmnr.opentelemetry_lib.opentelemetry.instrumentation.shared.base_instrumentor import BaseLaminarInstrumentor, LaminarInstrumentorConfig
from lmnr.opentelemetry_lib.opentelemetry.instrumentation.shared.types import WrappedFunctionSpec, LaminarInstrumentationScopeAttributes

from .wrappers import _wrap, _awrap

_instruments = ("daytona >= 0.1.0",)

class DaytonaSDKInstrumentor(BaseLaminarInstrumentor):
    _scope: LaminarInstrumentationScopeAttributes | None = None

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrumentation_scope(self) -> LaminarInstrumentationScopeAttributes:
        daytona_version = "unknown"
        try:
            daytona_version = version("daytona")
        except Exception:
            pass
        return LaminarInstrumentationScopeAttributes(
            name="daytona",
            version=daytona_version,
        )

    def instrumentation_scope(self) -> LaminarInstrumentationScopeAttributes:
        if self._scope is not None:
            return self._scope
        self._scope = self._instrumentation_scope()
        return self._scope

    def __init__(self):
        super().__init__()
        self.instrumentor_config =LaminarInstrumentorConfig(
            wrapped_functions= [
                WrappedFunctionSpec(
                    package_name="daytona._sync.process",
                    object_name="Process",
                    method_name="execute_session_command",
                    is_async=False,
                    is_streaming=False,
                    span_name="daytona.sandbox.process.execute_session_command",
                    span_type="DEFAULT",
                    instrumentation_scope=self.instrumentation_scope(),
                    wrapper_function=_wrap,
                ),
                WrappedFunctionSpec(
                    package_name="daytona._async.process",
                    object_name="AsyncProcess",
                    method_name="execute_session_command",
                    is_async=True,
                    is_streaming=False,
                    span_name="daytona.sandbox.process.execute_session_command",
                    span_type="DEFAULT",
                    instrumentation_scope=self.instrumentation_scope(),
                    wrapper_function=_awrap,
                ),
            ]
        )



import re
from pydantic import BaseModel
from opentelemetry.sdk.trace import Span
from opentelemetry.util.types import AttributeValue
from typing_extensions import TypedDict


class ToolDefinition(TypedDict):
    name: str | None
    description: str | None
    parameters: dict | None


def model_as_dict(model: BaseModel | dict) -> dict:
    if isinstance(model, BaseModel) and hasattr(model, "model_dump"):
        return model.model_dump()
    elif isinstance(model, dict):
        return model
    else:
        return dict(model)


def set_span_attribute(span: Span, key: str, value: AttributeValue | None):
    if value is None or value == "":
        return
    span.set_attribute(key, value)


def get_tool_definition(tool: dict) -> ToolDefinition:
    parameters = None
    description = None
    name = (tool.get("function") or {}).get("name") or tool.get("name")
    if tool.get("type") == "function":
        function = tool.get("function") or {}
        parameters = function.get("parameters") or tool.get("parameters")
        description = function.get("description") or tool.get("description")
    elif isinstance(tool.get("type"), str) and tool.get("type").startswith("computer"):
        # Anthropic beta computer tools
        # https://docs.anthropic.com/en/docs/agents-and-tools/tool-use/computer-use-tool

        # OpenAI computer use API
        # https://platform.openai.com/docs/guides/tools-computer-use
        if not name:
            name = tool.get("type")

        parameters = {}
        tool_parameters = (tool.get("function") or {}).get("parameters") or {}
        # Anthropic
        display_width_px = tool_parameters.get("display_width_px") or tool.get(
            "display_width_px"
        )
        display_height_px = tool_parameters.get("display_height_px") or tool.get(
            "display_height_px"
        )
        display_number = tool_parameters.get("display_number") or tool.get(
            "display_number"
        )
        if display_width_px:
            parameters["display_width_px"] = display_width_px
        if display_height_px:
            parameters["display_height_px"] = display_height_px
        if display_number:
            parameters["display_number"] = display_number
        # OpenAI
        display_width = tool_parameters.get("display_width") or tool.get(
            "display_width"
        )
        display_height = tool_parameters.get("display_height") or tool.get(
            "display_height"
        )
        environment = tool_parameters.get("environment") or tool.get("environment")
        if display_width:
            parameters["display_width"] = display_width
        if display_height:
            parameters["display_height"] = tool.get("display_height")
        if environment:  # Literal['browser', 'mac', 'windows', 'ubuntu']
            parameters["environment"] = environment
    else:
        # Some versions of LiteLLM (around 1.69.0) flatten the tool definition in
        # anthropic style, not the OpenAI style as they do with other tool types.
        function = tool.get("function") or tool
        parameters = function.get("parameters") or function.get("input_schema")
        description = function.get("description") or ""

    return ToolDefinition(
        name=name,
        description=description,
        parameters=parameters,
    )


def is_validator_iterator(content):
    """
    Some OpenAI objects contain fields typed as Iterable, which pydantic
    internally converts to a ValidatorIterator, and they cannot be trivially
    serialized without consuming the iterator to, for example, a list.

    See: https://github.com/pydantic/pydantic/issues/9541#issuecomment-2189045051
    """
    return re.search(r"pydantic.*ValidatorIterator'>$", str(type(content)))



import importlib
import sys

from lmnr.sdk.log import get_default_logger

from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.instrumentation.utils import unwrap
from typing import Any, Collection
from wrapt import FunctionWrapper, wrap_function_wrapper

from .wrappers import (
    wrap_sync,
    wrap_async,
    wrap_async_gen,
    wrap_transport_connect,
    wrap_transport_close,
    wrap_query,
    wrap_client_init,
)

logger = get_default_logger(__name__)

_instruments = ("claude-agent-sdk >= 0.1.0",)

WRAPPED_METHODS = [
    {
        "package": "claude_agent_sdk._internal.transport.subprocess_cli",
        "object": "SubprocessCLITransport",
        "method": "connect",
        "class_name": "SubprocessCLITransport",
        "wrapper": wrap_transport_connect,
    },
    {
        "package": "claude_agent_sdk._internal.transport.subprocess_cli",
        "object": "SubprocessCLITransport",
        "method": "close",
        "class_name": "SubprocessCLITransport",
        "wrapper": wrap_transport_close,
    },
    {
        "package": "claude_agent_sdk.client",
        "object": "ClaudeSDKClient",
        "method": "__init__",
        "class_name": "ClaudeSDKClient",
        "wrapper": wrap_client_init,
    },
    {
        "package": "claude_agent_sdk.client",
        "object": "ClaudeSDKClient",
        "method": "connect",
        "class_name": "ClaudeSDKClient",
        "should_publish_span_context": True,
        "wrapper": wrap_async,
    },
    {
        "package": "claude_agent_sdk.client",
        "object": "ClaudeSDKClient",
        "method": "query",
        "class_name": "ClaudeSDKClient",
        "should_publish_span_context": True,
        "wrapper": wrap_async,
    },
    {
        "package": "claude_agent_sdk.client",
        "object": "ClaudeSDKClient",
        "method": "receive_messages",
        "class_name": "ClaudeSDKClient",
        "wrapper": wrap_async_gen,
    },
    {
        "package": "claude_agent_sdk.client",
        "object": "ClaudeSDKClient",
        "method": "receive_response",
        "class_name": "ClaudeSDKClient",
        "wrapper": wrap_async_gen,
    },
    {
        "package": "claude_agent_sdk.client",
        "object": "ClaudeSDKClient",
        "method": "interrupt",
        "class_name": "ClaudeSDKClient",
        "wrapper": wrap_async,
    },
    {
        "package": "claude_agent_sdk.client",
        "object": "ClaudeSDKClient",
        "method": "disconnect",
        "class_name": "ClaudeSDKClient",
        "wrapper": wrap_async,
    },
    {
        # Module-level query function (streaming)
        "package": "claude_agent_sdk",
        "method": "query",
        "should_publish_span_context": True,
        "wrapper": wrap_query,
    },
    {
        # Module-level create_sdk_mcp_server function (sync)
        "package": "claude_agent_sdk",
        "method": "create_sdk_mcp_server",
        "wrapper": wrap_sync,
    },
]

_MODULE_FUNCTION_ORIGINALS: dict[tuple[str, str], Any] = {}


def _replace_function_aliases(original, wrapped):
    for module in list(sys.modules.values()):
        module_dict = getattr(module, "__dict__", None)
        if not module_dict:
            continue
        for attr, value in list(module_dict.items()):
            if value is original:
                setattr(module, attr, wrapped)


def _wrap_module_function(module_name: str, function_name: str, wrapper):
    try:
        module = sys.modules.get(module_name) or importlib.import_module(module_name)
    except ModuleNotFoundError:
        return

    try:
        original = getattr(module, function_name)
    except AttributeError:
        return

    key = (module_name, function_name)
    if key not in _MODULE_FUNCTION_ORIGINALS:
        _MODULE_FUNCTION_ORIGINALS[key] = original

    wrapped_function = FunctionWrapper(original, wrapper)
    setattr(module, function_name, wrapped_function)
    _replace_function_aliases(original, wrapped_function)


def _unwrap_module_function(module_name: str, function_name: str):
    key = (module_name, function_name)
    original = _MODULE_FUNCTION_ORIGINALS.get(key)
    if not original:
        return

    module = sys.modules.get(module_name)
    if not module:
        return

    current = getattr(module, function_name, None)
    setattr(module, function_name, original)
    if current is not None:
        _replace_function_aliases(current, original)
    del _MODULE_FUNCTION_ORIGINALS[key]


class ClaudeAgentInstrumentor(BaseInstrumentor):
    def __init__(self):
        super().__init__()

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        # Wrap methods
        for wrapped_method in WRAPPED_METHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            wrap_method = wrapped_method.get("method")
            wrapper_func = wrapped_method.get("wrapper")

            # Create wrapper instance with metadata
            wrapper = wrapper_func(wrapped_method)

            if wrap_object:
                target = f"{wrap_object}.{wrap_method}"
                try:
                    wrap_function_wrapper(
                        wrap_package,
                        target,
                        wrapper,
                    )
                except (ModuleNotFoundError, AttributeError):
                    pass  # that's ok, we don't want to fail if some methods do not exist
            else:
                try:
                    _wrap_module_function(
                        wrap_package,
                        wrap_method,
                        wrapper,
                    )
                except (ModuleNotFoundError, AttributeError):
                    pass  # that's ok

    def _uninstrument(self, **kwargs):
        for wrapped_method in WRAPPED_METHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            wrap_method = wrapped_method.get("method")

            if wrap_object:
                module_path = f"{wrap_package}.{wrap_object}"
                try:
                    unwrap(module_path, wrap_method)
                except (ModuleNotFoundError, AttributeError):
                    pass  # that's ok, we don't want to fail if some methods do not exist
            else:
                _unwrap_module_function(wrap_package, wrap_method)



from abc import ABC, abstractmethod
from pathlib import Path

import uuid

from lmnr.sdk.client.synchronous.sync_client import LaminarClient
from lmnr.sdk.datasets.file_utils import load_from_paths
from lmnr.sdk.log import get_default_logger
from lmnr.sdk.types import Datapoint

DEFAULT_FETCH_SIZE = 25
LOG = get_default_logger(__name__, verbose=False)


class EvaluationDataset(ABC):
    @abstractmethod
    def __init__(self, *args, **kwargs):
        pass

    @abstractmethod
    def __len__(self) -> int:
        pass

    @abstractmethod
    def __getitem__(self, idx) -> Datapoint:
        pass

    def slice(self, start: int, end: int):
        return [self[i] for i in range(max(start, 0), min(end, len(self)))]


class LaminarDataset(EvaluationDataset):
    client: LaminarClient
    id: uuid.UUID | None = None

    def __init__(
        self,
        name: str | None = None,
        id: uuid.UUID | None = None,
        fetch_size: int = DEFAULT_FETCH_SIZE,
    ):
        self.name = name
        self.id = id
        if name is None and id is None:
            raise ValueError("Either name or id must be provided")
        if name is not None and id is not None:
            raise ValueError("Only one of name or id must be provided")
        self._len = None
        self._fetched_items = []
        self._offset = 0
        self._fetch_size = fetch_size
        self._logger = get_default_logger(self.__class__.__name__)

    def _fetch_batch(self):
        self._logger.debug(
            f"dataset name: {self.name}, id: {self.id}. Fetching batch from {self._offset} to "
            + f"{self._offset + self._fetch_size}"
        )
        identifier = {"id": self.id} if self.id is not None else {"name": self.name}
        resp = self.client.datasets.pull(
            **identifier,
            offset=self._offset,
            limit=self._fetch_size,
        )
        self._fetched_items += resp.items
        self._offset = len(self._fetched_items)
        if self._len is None:
            self._len = resp.total_count

    def __len__(self) -> int:
        if self._len is None:
            self._fetch_batch()
        return self._len

    def __getitem__(self, idx) -> Datapoint:
        if idx >= len(self._fetched_items):
            self._fetch_batch()
        return self._fetched_items[idx]

    def set_client(self, client: LaminarClient):
        self.client = client

    def push(self, paths: str | list[str], recursive: bool = False):
        paths = [paths] if isinstance(paths, str) else paths
        paths = [Path(path) for path in paths]
        data = load_from_paths(paths, recursive)
        if len(data) == 0:
            LOG.warning("No data to push. Skipping")
            return
        identifier = {"id": self.id} if self.id is not None else {"name": self.name}
        self.client.datasets.push(data, **identifier)
        LOG.info(
            f"Successfully pushed {len(data)} datapoints to dataset [{identifier}]"
        )



import asyncio
import logging
import inspect
import os
import threading
import traceback
from contextlib import asynccontextmanager
from importlib.metadata import version
from packaging.version import parse

from opentelemetry import context as context_api
from opentelemetry._events import EventLogger
from .shared.config import Config

import openai


_OPENAI_VERSION = version("openai")

LMNR_TRACE_CONTENT = "LMNR_TRACE_CONTENT"


def is_openai_v1():
    return parse(_OPENAI_VERSION) >= parse("1.0.0")


def is_reasoning_supported():
    # Reasoning has been introduced in OpenAI API on Dec 17, 2024
    #     as per https://platform.openai.com/docs/changelog.
    # The updated OpenAI library version is 1.58.0
    #     as per https://pypi.org/project/openai/.
    return parse(_OPENAI_VERSION) >= parse("1.58.0")


def is_azure_openai(instance):

    return is_openai_v1() and isinstance(
        instance._client, (openai.AsyncAzureOpenAI, openai.AzureOpenAI)
    )


def is_metrics_enabled() -> bool:
    return False


def should_record_stream_token_usage():
    return Config.enrich_token_usage


def _with_image_gen_metric_wrapper(func):
    def _with_metric(duration_histogram, exception_counter):
        def wrapper(wrapped, instance, args, kwargs):
            return func(
                duration_histogram,
                exception_counter,
                wrapped,
                instance,
                args,
                kwargs,
            )

        return wrapper

    return _with_metric


def _with_embeddings_telemetry_wrapper(func):
    def _with_embeddings_telemetry(
        tracer,
        token_counter,
        vector_size_counter,
        duration_histogram,
        exception_counter,
    ):
        def wrapper(wrapped, instance, args, kwargs):
            return func(
                tracer,
                token_counter,
                vector_size_counter,
                duration_histogram,
                exception_counter,
                wrapped,
                instance,
                args,
                kwargs,
            )

        return wrapper

    return _with_embeddings_telemetry


def _with_chat_telemetry_wrapper(func):
    def _with_chat_telemetry(
        tracer,
        token_counter,
        choice_counter,
        duration_histogram,
        exception_counter,
        streaming_time_to_first_token,
        streaming_time_to_generate,
    ):
        def wrapper(wrapped, instance, args, kwargs):
            return func(
                tracer,
                token_counter,
                choice_counter,
                duration_histogram,
                exception_counter,
                streaming_time_to_first_token,
                streaming_time_to_generate,
                wrapped,
                instance,
                args,
                kwargs,
            )

        return wrapper

    return _with_chat_telemetry


def _with_tracer_wrapper(func):
    def _with_tracer(tracer):
        def wrapper(wrapped, instance, args, kwargs):
            return func(tracer, wrapped, instance, args, kwargs)

        return wrapper

    return _with_tracer


@asynccontextmanager
async def start_as_current_span_async(tracer, *args, **kwargs):
    with tracer.start_as_current_span(*args, **kwargs) as span:
        yield span


def dont_throw(func):
    """
    A decorator that wraps the passed in function and logs exceptions instead of throwing them.
    Works for both synchronous and asynchronous functions.
    """
    logger = logging.getLogger(func.__module__)

    async def async_wrapper(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except Exception as e:
            _handle_exception(e, func, logger)

    def sync_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            _handle_exception(e, func, logger)

    def _handle_exception(e, func, logger):
        logger.debug(
            "OpenLLMetry failed to trace in %s, error: %s",
            func.__name__,
            traceback.format_exc(),
        )
        if Config.exception_logger:
            Config.exception_logger(e)

    return async_wrapper if inspect.iscoroutinefunction(func) else sync_wrapper


def run_async(method):
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = None

    if loop and loop.is_running():
        thread = threading.Thread(target=lambda: asyncio.run(method))
        thread.start()
        thread.join()
    else:
        asyncio.run(method)


def should_send_prompts():
    return (
        os.getenv(LMNR_TRACE_CONTENT) or "true"
    ).lower() == "true" or context_api.get_value("override_enable_content_tracing")


def should_emit_events() -> bool:
    """
    Checks if the instrumentation isn't using the legacy attributes
    and if the event logger is not None.
    """
    return not Config.use_legacy_attributes and isinstance(
        Config.event_logger, EventLogger
    )



"""Rollout session management resource for synchronous client."""

from typing import ContextManager, Literal
import httpx
import uuid

from lmnr.sdk.client.synchronous.resources.base import BaseResource
from lmnr.sdk.client.types import SpanStartData
from lmnr.sdk.log import get_default_logger
from lmnr.sdk.types import RolloutParam

logger = get_default_logger(__name__)


class Rollout(BaseResource):
    """
    Rollout session management resource.

    Provides methods for connecting to rollout sessions, streaming spans,
    and managing rollout lifecycle.
    """

    def connect(
        self,
        session_id: uuid.UUID,
        function_name: str,
        params: list[RolloutParam] | None = None,
    ) -> ContextManager[httpx.Response]:
        """
        Connect to a rollout session and return the SSE streaming response.

        This creates the session and returns a streaming response that should
        be parsed for SSE events.

        Args:
            session_id: UUID for the rollout session
            function_name: Name of the function being rolled out
            params: List of parameter metadata for the function

        Returns:
            httpx.Response: Streaming response for SSE events

        Raises:
            httpx.HTTPStatusError: If the request fails
        """
        payload = {
            "name": function_name,
            "params": params or [],
        }

        headers = self._headers()
        headers["Accept"] = "text/event-stream"

        response = self._client.stream(
            "POST",
            f"{self._base_url}/v1/rollouts/{session_id}",
            headers=headers,
            json=payload,
            timeout=None,  # No timeout for streaming connection
        )
        return response

    def update_span_info(
        self,
        session_id: str,
        span_data: SpanStartData,
    ) -> None:
        """
        Stream a span update in real-time to the backend.

        This is used during rollout execution to provide live updates
        in the Laminar UI.

        Args:
            session_id: UUID of the rollout session
            span_data: Span data to stream

        Raises:
            httpx.HTTPStatusError: If the request fails
        """

        try:
            response = self._client.patch(
                f"{self._base_url}/v1/rollouts/{session_id}/update",
                headers=self._headers(),
                json={
                    "type": "spanStart",
                    "name": span_data["name"],
                    "spanId": str(span_data["span_id"]),
                    "parentSpanId": (
                        str(span_data["parent_span_id"])
                        if span_data["parent_span_id"]
                        else None
                    ),
                    "traceId": str(span_data["trace_id"]),
                    "startTime": span_data["start_time"].isoformat(),
                    "attributes": span_data["attributes"],
                    "spanType": span_data["span_type"],
                },
                timeout=5.0,  # Short timeout for streaming
            )
            response.raise_for_status()
        except Exception as e:
            # Log but don't raise - streaming failures shouldn't break execution
            logger.debug(f"Failed to update span info for session {session_id}: {e}")

    def delete_session(
        self,
        session_id: str,
    ) -> None:
        """
        Delete a rollout session.

        Args:
            session_id: UUID of the rollout session to delete

        Raises:
            httpx.HTTPStatusError: If the request fails
        """
        response = self._client.delete(
            f"{self._base_url}/v1/rollouts/{session_id}",
            headers=self._headers(),
        )
        response.raise_for_status()

    def update_status(
        self,
        session_id: str,
        status: Literal["PENDING", "RUNNING", "FINISHED", "STOPPED"],
    ) -> None:
        """
        Update the status of a rollout session.

        Args:
            session_id: UUID of the rollout session
            status: New status (e.g., 'PENDING', 'RUNNING', 'FINISHED', 'STOPPED')

        Raises:
            httpx.HTTPStatusError: If the request fails
        """
        response = self._client.patch(
            f"{self._base_url}/v1/rollouts/{session_id}/status",
            headers=self._headers(),
            json={"status": status},
        )
        response.raise_for_status()



from typing import Any, Sequence

from opentelemetry.trace import Span
from pydantic import BaseModel

from lmnr.sdk.utils import json_dumps
from lmnr.opentelemetry_lib.opentelemetry.instrumentation.shared.utils import (
    dont_throw,
    extract_json_schema,
    set_span_attribute,
    to_dict,
)
from ...utils import infer_provider


@dont_throw
def process_completion_kwargs(
    span: Span,
    args: Sequence[Any] | None = None,
    kwargs: dict[str, Any] | None = None,
):
    if kwargs is None:
        kwargs = {}
    model = args[0] if args else kwargs.get("model")
    set_span_attribute(span, "gen_ai.request.model", model)
    set_span_attribute(span, "gen_ai.system", infer_provider(model))
    set_span_attribute(span, "gen_ai.request.temperature", kwargs.get("temperature"))
    set_span_attribute(span, "gen_ai.request.top_p", kwargs.get("top_p"))
    set_span_attribute(span, "gen_ai.request.top_k", kwargs.get("top_logprobs"))
    set_span_attribute(
        span,
        "gen_ai.request.max_tokens",
        kwargs.get("max_tokens", kwargs.get("max_completion_tokens")),
    )
    set_span_attribute(span, "gen_ai.request.stop_sequences", kwargs.get("stop"))
    set_span_attribute(
        span, "gen_ai.request.frequency_penalty", kwargs.get("frequency_penalty")
    )
    set_span_attribute(
        span, "gen_ai.request.presence_penalty", kwargs.get("presence_penalty")
    )
    set_span_attribute(
        span, "gen_ai.request.reasoning_effort", kwargs.get("reasoning_effort")
    )
    set_span_attribute(
        span,
        "gen_ai.request.thinking_budget",
        (kwargs.get("thinking") or {}).get("budget_tokens"),
    )
    set_span_attribute(
        span,
        "gen_ai.request.base_url",
        kwargs.get("base_url"),
    )
    if kwargs.get("response_format"):
        set_span_attribute(
            span,
            "gen_ai.request.structured_output_schema",
            json_dumps(extract_json_schema(kwargs.get("response_format"))),
        )


@dont_throw
def process_completion_inputs(
    span: Span,
    messages: list[dict],
    tools: list[dict] | None = None,
):
    # this for loop replicates `litellm.utils.validate_and_fix_openai_messages`
    attr_messages = []
    for msg in messages:
        message = to_dict(msg)
        if not message.get("role"):
            message["role"] = "assistant"
        if isinstance(message.get("tool_calls"), list):
            message["tool_calls"] = [
                to_dict(tool_call) for tool_call in message.get("tool_calls")
            ]
        attr_messages.append(message)

    span.set_attribute("gen_ai.input.messages", json_dumps(attr_messages))

    if tools:
        attr_tools = [to_dict(tool) for tool in tools]
        span.set_attribute("gen_ai.tool.definitions", json_dumps(attr_tools))


@dont_throw
def process_completion_response(
    span: Span,
    response: BaseModel,
    record_raw_response: bool = False,
):
    response_dict = to_dict(response)
    set_span_attribute(span, "gen_ai.response.id", response_dict.get("id"))
    set_span_attribute(span, "gen_ai.response.model", response_dict.get("model"))
    set_span_attribute(
        span,
        "gen_ai.response.system_fingerprint",
        response_dict.get("system_fingerprint"),
    )
    choices = response_dict.get("choices", [])
    messages = [choice.get("message", {}) for choice in choices]
    span.set_attribute("gen_ai.output.messages", json_dumps(messages))
    usage = to_dict(response_dict.get("usage", {}))
    input_tokens = usage.get("prompt_tokens", usage.get("input_tokens", 0))
    output_tokens = usage.get("completion_tokens", usage.get("output_tokens", 0))
    total_tokens = usage.get("total_tokens", input_tokens + output_tokens)
    set_span_attribute(span, "gen_ai.usage.input_tokens", input_tokens)
    set_span_attribute(span, "gen_ai.usage.output_tokens", output_tokens)
    set_span_attribute(span, "llm.usage.total_tokens", total_tokens)
    if input_details := usage.get("prompt_tokens_details"):
        details = to_dict(input_details)
        cache_read_tokens = details.get(
            "cached_tokens", usage.get("cache_read_input_tokens", 0)
        )
        cache_creation_tokens = details.get(
            "cache_creation_tokens", usage.get("cache_creation_input_tokens", 0)
        )
        set_span_attribute(
            span, "gen_ai.usage.cache_read_input_tokens", cache_read_tokens
        )
        set_span_attribute(
            span, "gen_ai.usage.cache_creation_input_tokens", cache_creation_tokens
        )

    # Record raw response in rollout mode
    if record_raw_response:
        set_span_attribute(span, "lmnr.sdk.raw.response", json_dumps(response_dict))

    return response



"""
Rollout wrapper for LiteLLM instrumentation.

This module adds caching and override capabilities to LiteLLM instrumentation
during rollout sessions.
"""

import json
from typing import Any, AsyncGenerator, Generator

from lmnr.sdk.laminar import Laminar
from lmnr.sdk.log import get_default_logger
from lmnr.sdk.rollout.instrumentation import RolloutInstrumentationWrapper
from lmnr.sdk.rollout_control import is_rollout_mode

logger = get_default_logger(__name__)


class DualIteratorWrapper:
    """
    Wrapper that supports both sync and async iteration.
    This allows cached streaming responses to work in both contexts.
    """

    def __init__(self, sync_iterator, cached_response=None):
        """
        Args:
            sync_iterator: A sync iterator/generator to wrap
            cached_response: Optional cached response object for setting span attributes
        """
        self._sync_iterator = sync_iterator
        self._items = None  # Cache items for potential reuse
        self._cached_response = cached_response

    def set_span_attributes(self, span, record_raw_response=False):
        """
        Set span attributes from the cached response without consuming the iterator.

        Args:
            span: The span to set attributes on
            record_raw_response: Whether to record raw response
        """
        if not self._cached_response:
            return

        from lmnr.sdk.utils import json_dumps
        from lmnr.opentelemetry_lib.opentelemetry.instrumentation.shared.utils import (
            set_span_attribute,
            to_dict,
        )

        response_dict = to_dict(self._cached_response)

        # Set common attributes based on response type
        set_span_attribute(span, "gen_ai.response.id", response_dict.get("id"))
        set_span_attribute(span, "gen_ai.response.model", response_dict.get("model"))

        # For completion responses
        if "choices" in response_dict:
            set_span_attribute(
                span,
                "gen_ai.response.system_fingerprint",
                response_dict.get("system_fingerprint"),
            )
            messages = [
                choice.get("message", {}) for choice in response_dict.get("choices", [])
            ]
            span.set_attribute("gen_ai.output.messages", json_dumps(messages))

        # For responses API
        elif "output" in response_dict:
            final_items = []
            if reasoning := response_dict.get("reasoning"):
                reasoning_dict = to_dict(reasoning)
                if reasoning_dict.get("summary") or reasoning_dict.get("effort"):
                    final_items.append(reasoning_dict)
            if isinstance(response_dict.get("output"), list):
                for item in response_dict.get("output"):
                    final_items.append(to_dict(item))
            span.set_attribute("gen_ai.output.messages", json_dumps(final_items))

        # Record raw response in rollout mode
        if record_raw_response:
            set_span_attribute(span, "lmnr.sdk.raw.response", json_dumps(response_dict))

    def __iter__(self):
        """Sync iteration support."""
        if self._items is not None:
            # If we already consumed it, replay from cache
            for item in self._items:
                yield item
            return

        # First time iteration - consume and cache
        self._items = []
        for item in self._sync_iterator:
            self._items.append(item)
            yield item

    async def __aiter__(self):
        """Async iteration support."""
        if self._items is not None:
            # If we already consumed it, replay from cache
            for item in self._items:
                yield item
        else:
            # First time iteration - consume and cache
            self._items = []
            for item in self._sync_iterator:
                self._items.append(item)
                yield item


def _import_litellm_types():
    """Lazy import litellm types to avoid import errors."""
    types = {
        "ModelResponse": None,
        "ResponsesAPIResponse": None,
        "Delta": None,
        "ResponseCompletedEvent": None,
    }

    try:
        from litellm import ModelResponse

        types["ModelResponse"] = ModelResponse
    except ImportError:
        pass

    try:
        from litellm.responses.main import ResponsesAPIResponse

        types["ResponsesAPIResponse"] = ResponsesAPIResponse
    except ImportError:
        pass

    try:
        from litellm.types.utils import Delta

        types["Delta"] = Delta
    except ImportError:
        pass

    try:
        from litellm.types.llms.openai import ResponseCompletedEvent

        types["ResponseCompletedEvent"] = ResponseCompletedEvent
    except ImportError:
        pass

    return types


class LiteLLMRolloutWrapper(RolloutInstrumentationWrapper):
    """
    Rollout wrapper specific to LiteLLM instrumentation.

    Handles:
    - Converting cached responses to LiteLLM format (both completion and responses)
    - Applying overrides to system prompts and tool definitions
    - Setting rollout-specific span attributes
    """

    def cached_response_to_completion(self, cached_span: dict[str, Any]) -> Any:
        """
        Convert cached span data to LiteLLM ModelResponse format.

        Args:
            cached_span: Cached span data from cache server

        Returns:
            ModelResponse object, dict, or None
        """
        types = _import_litellm_types()
        ModelResponse = types["ModelResponse"]

        # Try to parse from raw response first
        if raw_response := cached_span.get("attributes", {}).get(
            "lmnr.sdk.raw.response"
        ):
            try:
                response_dict = None
                if isinstance(raw_response, dict):
                    response_dict = raw_response
                elif isinstance(raw_response, str):
                    response_dict = json.loads(raw_response)

                if response_dict and ModelResponse:
                    try:
                        return ModelResponse.model_validate(response_dict)
                    except Exception as e:
                        logger.debug(
                            f"Failed to validate ModelResponse, returning dict: {e}"
                        )
                        return response_dict
                elif response_dict:
                    return response_dict
            except Exception as e:
                logger.debug(f"Failed to parse raw response: {e}")
                pass  # fallback to the legacy parsing path

        try:
            attributes = cached_span.get("attributes", {})
            output_str = cached_span.get("output", "")

            if not output_str:
                logger.warning("Cached span has no output")
                return None

            # Parse the output JSON - should be array of messages
            messages = json.loads(output_str)
            if not isinstance(messages, list):
                logger.warning(f"Unexpected output format: {type(messages)}")
                return None

            # Reconstruct ModelResponse dict
            choices = []
            for i, message in enumerate(messages):
                choices.append(
                    {
                        "index": i,
                        "message": message,
                        "finish_reason": "stop",  # Default for cached responses
                    }
                )

            response_dict = {
                "id": attributes.get("gen_ai.response.id", "cached"),
                "model": attributes.get("gen_ai.response.model", "unknown"),
                "choices": choices,
                "created": int(cached_span.get("start_time", 0) / 1_000_000_000),
                "object": "chat.completion",
                "system_fingerprint": attributes.get(
                    "gen_ai.response.system_fingerprint"
                ),
                "usage": None,  # Cached responses don't track tokens
            }

            # Try to construct ModelResponse object
            if ModelResponse:
                try:
                    return ModelResponse.model_validate(response_dict)
                except Exception as e:
                    logger.debug(
                        f"Failed to validate ModelResponse, returning dict: {e}"
                    )
                    return response_dict

            return response_dict

        except Exception as e:
            logger.debug(
                f"Failed to convert cached response to completion format: {e}",
                exc_info=True,
            )
            return None

    def cached_response_to_responses(self, cached_span: dict[str, Any]) -> Any:
        """
        Convert cached span data to LiteLLM ResponsesAPIResponse format.

        Args:
            cached_span: Cached span data from cache server

        Returns:
            ResponsesAPIResponse object, dict, or None
        """
        types = _import_litellm_types()
        ResponsesAPIResponse = types["ResponsesAPIResponse"]

        # Try to parse from raw response first
        if raw_response := cached_span.get("attributes", {}).get(
            "lmnr.sdk.raw.response"
        ):
            try:
                response_dict = None
                if isinstance(raw_response, dict):
                    response_dict = raw_response
                elif isinstance(raw_response, str):
                    response_dict = json.loads(raw_response)

                if response_dict and ResponsesAPIResponse:
                    try:
                        return ResponsesAPIResponse.model_validate(response_dict)
                    except Exception as e:
                        logger.debug(
                            f"Failed to validate ResponsesAPIResponse, returning dict: {e}"
                        )
                        return response_dict
                elif response_dict:
                    return response_dict
            except Exception as e:
                logger.debug(f"Failed to parse raw response: {e}")
                pass  # fallback to the legacy parsing path

        try:
            attributes = cached_span.get("attributes", {})
            output_str = cached_span.get("output", "")

            if not output_str:
                logger.warning("Cached span has no output")
                return None

            # Parse the output JSON - may contain reasoning + output items
            items = json.loads(output_str)
            if not isinstance(items, list):
                logger.warning(f"Unexpected output format: {type(items)}")
                return None

            # Separate reasoning from output items
            reasoning = None
            output_items = []

            for item in items:
                # Check if it's a reasoning object (has summary or effort)
                if isinstance(item, dict) and (
                    item.get("summary") or item.get("effort")
                ):
                    reasoning = item
                else:
                    output_items.append(item)

            # Reconstruct ResponsesAPIResponse dict
            response_dict = {
                "id": attributes.get("gen_ai.response.id", "cached"),
                "model": attributes.get("gen_ai.response.model", "unknown"),
                "output": output_items,
                "usage": None,  # Cached responses don't track tokens
                "created_at": int(cached_span.get("end_time", 0) / 1_000_000_000),
            }

            if reasoning:
                response_dict["reasoning"] = reasoning

            # Try to construct ResponsesAPIResponse object
            if ResponsesAPIResponse:
                try:
                    return ResponsesAPIResponse.model_validate(response_dict)
                except Exception as e:
                    logger.debug(
                        f"Failed to validate ResponsesAPIResponse, returning dict: {e}"
                    )
                    return response_dict

            return response_dict

        except Exception as e:
            logger.debug(
                f"Failed to convert cached response to responses format: {e}",
                exc_info=True,
            )
            return None

    def _create_cached_completion_stream(
        self, response: Any
    ) -> Generator[Any, None, None]:
        """
        Create a generator that yields a cached completion response as a single chunk.

        Args:
            response: Cached response to yield (ModelResponse or dict)

        Yields:
            Chunk object or dict: The cached response formatted as a chunk
        """
        types = _import_litellm_types()
        ModelResponse = types["ModelResponse"]
        Delta = types["Delta"]

        # Convert response to dict if needed
        if hasattr(response, "model_dump"):
            response_dict = response.model_dump()
        else:
            response_dict = response

        # Convert the full response to a streaming chunk format
        chunk_dict = {
            "id": response_dict.get("id"),
            "model": response_dict.get("model"),
            "object": "chat.completion.chunk",
            "created": response_dict.get("created"),
            "choices": [],
        }

        for choice in response_dict.get("choices", []):
            chunk_dict["choices"].append(
                {
                    "index": choice.get("index", 0),
                    "delta": choice.get("message", {}),
                    "finish_reason": choice.get("finish_reason"),
                }
            )

        # Try to validate as ModelResponse if available
        if ModelResponse:
            try:
                chunk = ModelResponse.model_validate(chunk_dict)
                if Delta:
                    for choice in chunk.choices:
                        if hasattr(choice, "delta"):
                            choice.delta = Delta.model_validate(choice.delta)
                yield chunk
                return
            except Exception:
                logger.debug("Failed to validate ModelResponse", exc_info=True)
                pass

        yield chunk_dict

    async def _create_async_cached_completion_stream(
        self, response: Any
    ) -> AsyncGenerator[Any, None]:
        """
        Create an async generator that yields a cached completion response as a single chunk.

        Args:
            response: Cached response to yield (ModelResponse or dict)

        Yields:
            Chunk object or dict: The cached response formatted as a chunk
        """
        types = _import_litellm_types()
        ModelResponse = types["ModelResponse"]
        Delta = types["Delta"]

        # Convert response to dict if needed
        if hasattr(response, "model_dump"):
            response_dict = response.model_dump()
        else:
            response_dict = response

        # Convert the full response to a streaming chunk format
        chunk_dict = {
            "id": response_dict.get("id"),
            "model": response_dict.get("model"),
            "object": "chat.completion.chunk",
            "created": response_dict.get("created"),
            "choices": [],
        }

        for choice in response_dict.get("choices", []):
            chunk_dict["choices"].append(
                {
                    "index": choice.get("index", 0),
                    "delta": choice.get("message", {}),
                    "finish_reason": choice.get("finish_reason"),
                }
            )

        # Try to validate as ModelResponse if available
        if ModelResponse:
            try:
                chunk = ModelResponse.model_validate(chunk_dict)
                if Delta:
                    for choice in chunk.choices:
                        if hasattr(choice, "delta"):
                            choice.delta = Delta.model_validate(choice.delta)
                yield chunk
                return
            except Exception:
                pass

        yield chunk_dict

    def _create_cached_responses_stream(
        self, response: Any
    ) -> Generator[Any, None, None]:
        """
        Create a generator that yields a cached responses response as a completed event.

        Args:
            response: Cached response to yield (ResponsesAPIResponse or dict)

        Yields:
            ResponseCompletedEvent or dict: The cached response formatted as a response.completed event
        """
        types = _import_litellm_types()
        ResponseCompletedEvent = types["ResponseCompletedEvent"]

        event_dict = {
            "type": "response.completed",
            "response": response,
        }

        # Try to validate as ResponseCompletedEvent if available
        if ResponseCompletedEvent:
            try:
                event = ResponseCompletedEvent.model_validate(event_dict)
                yield event
                return
            except Exception:
                logger.debug("Failed to validate ResponseCompletedEvent", exc_info=True)
                pass

        yield event_dict

    async def _create_async_cached_responses_stream(
        self, response: Any
    ) -> AsyncGenerator[Any, None]:
        """
        Create an async generator that yields a cached responses response as a completed event.

        Args:
            response: Cached response to yield (ResponsesAPIResponse or dict)

        Yields:
            ResponseCompletedEvent or dict: The cached response formatted as a response.completed event
        """
        types = _import_litellm_types()
        ResponseCompletedEvent = types["ResponseCompletedEvent"]

        event_dict = {
            "type": "response.completed",
            "response": response,
        }

        # Try to validate as ResponseCompletedEvent if available
        if ResponseCompletedEvent:
            try:
                event = ResponseCompletedEvent.model_validate(event_dict)
                yield event
                return
            except Exception:
                logger.debug("Failed to validate ResponseCompletedEvent", exc_info=True)
                pass

        yield event_dict

    def wrap_completion(
        self,
        wrapped,
        args,
        kwargs,
        is_streaming: bool = False,
    ) -> Any:
        """
        Wrap completion with rollout logic.

        Args:
            wrapped: Original function
            instance: Instance object
            args: Positional arguments
            kwargs: Keyword arguments
            is_streaming: Whether this is a streaming call
            is_async: Whether this is an async call

        Returns:
            Response (cached or live), or Generator/AsyncGenerator
        """
        # Check if rollout mode is active
        if not self.should_use_rollout():
            return wrapped(*args, **kwargs)

        # Get span path
        span_path = self.get_span_path()
        if not span_path:
            return wrapped(*args, **kwargs)

        # Get current call index
        current_index = self.get_current_index_for_path(span_path)

        logger.debug(f"LiteLLM completion call at {span_path}:{current_index}")
        span = Laminar.get_current_span()

        # Check cache
        if self.should_use_cache(span_path, current_index):
            cached_span = self.get_cached_response(span_path, current_index)
            if cached_span:
                logger.debug(
                    f"Using cached response for LiteLLM completion at {span_path}:{current_index}"
                )

                # Convert cached data to completion response
                response = self.cached_response_to_completion(cached_span)
                if response:
                    # Mark span as cached
                    try:
                        if span and span.is_recording():
                            span.set_attributes(
                                {
                                    "lmnr.span.type": "CACHED",
                                    "lmnr.rollout.cache_index": current_index,
                                    "lmnr.span.original_type": "LLM",
                                }
                            )
                    except Exception:
                        pass

                    # For streaming, return a dual iterator that works in both sync and async contexts
                    if is_streaming:
                        sync_gen = self._create_cached_completion_stream(response)
                        return DualIteratorWrapper(sync_gen, cached_response=response)

                    return response

        # Cache miss or conversion failed - execute live
        logger.debug(
            f"Executing live LiteLLM completion call for {span_path}:{current_index}"
        )
        return wrapped(*args, **kwargs)

    def wrap_responses(
        self,
        wrapped,
        args,
        kwargs,
        is_streaming: bool = False,
    ) -> Any:
        """
        Wrap responses with rollout logic.

        Args:
            wrapped: Original function
            args: Positional arguments
            kwargs: Keyword arguments
            is_streaming: Whether this is a streaming call

        Returns:
            Response (cached or live), or Generator/AsyncGenerator
        """
        # Check if rollout mode is active
        if not self.should_use_rollout():
            return wrapped(*args, **kwargs)

        # Get span path
        span_path = self.get_span_path()
        if not span_path:
            return wrapped(*args, **kwargs)

        # Get current call index
        current_index = self.get_current_index_for_path(span_path)

        logger.debug(f"LiteLLM responses call at {span_path}:{current_index}")
        span = Laminar.get_current_span()

        # Check cache
        if self.should_use_cache(span_path, current_index):
            cached_span = self.get_cached_response(span_path, current_index)
            if cached_span:
                logger.debug(
                    f"Using cached response for LiteLLM responses at {span_path}:{current_index}"
                )

                # Convert cached data to responses response
                response = self.cached_response_to_responses(cached_span)
                if response:
                    # Mark span as cached
                    try:
                        if span and span.is_recording():
                            span.set_attributes(
                                {
                                    "lmnr.span.type": "CACHED",
                                    "lmnr.rollout.cache_index": current_index,
                                    "lmnr.span.original_type": "LLM",
                                }
                            )
                    except Exception:
                        pass

                    # For streaming, return a dual iterator that works in both sync and async contexts
                    if is_streaming:
                        sync_gen = self._create_cached_responses_stream(response)
                        return DualIteratorWrapper(sync_gen, cached_response=response)

                    return response

        # Cache miss or conversion failed - execute live
        logger.debug(
            f"Executing live LiteLLM responses call for {span_path}:{current_index}"
        )
        return wrapped(*args, **kwargs)


# Singleton instance
_litellm_rollout_wrapper: LiteLLMRolloutWrapper | None = None


def get_litellm_rollout_wrapper() -> LiteLLMRolloutWrapper | None:
    """
    Get or create the LiteLLM rollout wrapper singleton.

    Returns:
        Optional[LiteLLMRolloutWrapper]: Wrapper instance or None if not in rollout mode
    """
    global _litellm_rollout_wrapper

    if not is_rollout_mode():
        return None

    if _litellm_rollout_wrapper is None:
        try:
            _litellm_rollout_wrapper = LiteLLMRolloutWrapper()
        except Exception as e:
            logger.error(f"Failed to create LiteLLM rollout wrapper: {e}")
            return None

    return _litellm_rollout_wrapper



"""OpenTelemetry Kernel instrumentation"""

import functools
from typing import Collection

from lmnr.opentelemetry_lib.opentelemetry.instrumentation.kernel.utils import (
    process_tool_output_formatter,
    screenshot_tool_output_formatter,
)
from lmnr.sdk.decorators import observe
from lmnr.sdk.utils import get_input_from_func_args, is_async, json_dumps
from lmnr import Laminar
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.instrumentation.utils import unwrap

from opentelemetry.trace.status import Status, StatusCode
from wrapt import wrap_function_wrapper

_instruments = ("kernel >= 0.2.0",)


WRAPPED_METHODS = [
    {
        "package": "kernel.resources.browsers",
        "object": "BrowsersResource",
        "method": "create",
        "class_name": "Browser",
    },
    {
        "package": "kernel.resources.browsers",
        "object": "BrowsersResource",
        "method": "retrieve",
        "class_name": "Browser",
    },
    {
        "package": "kernel.resources.browsers",
        "object": "BrowsersResource",
        "method": "list",
        "class_name": "Browser",
    },
    {
        "package": "kernel.resources.browsers",
        "object": "BrowsersResource",
        "method": "delete",
        "class_name": "Browser",
    },
    {
        "package": "kernel.resources.browsers",
        "object": "BrowsersResource",
        "method": "delete_by_id",
        "class_name": "Browser",
    },
    {
        "package": "kernel.resources.browsers",
        "object": "BrowsersResource",
        "method": "load_extensions",
        "class_name": "Browser",
    },
    {
        "package": "kernel.resources.browsers.computer",
        "object": "ComputerResource",
        "method": "capture_screenshot",
        "class_name": "Computer",
        "span_type": "TOOL",
        "output_formatter": screenshot_tool_output_formatter,
    },
    {
        "package": "kernel.resources.browsers.computer",
        "object": "ComputerResource",
        "method": "click_mouse",
        "class_name": "Computer",
        "span_type": "TOOL",
    },
    {
        "package": "kernel.resources.browsers.computer",
        "object": "ComputerResource",
        "method": "drag_mouse",
        "class_name": "Computer",
        "span_type": "TOOL",
    },
    {
        "package": "kernel.resources.browsers.computer",
        "object": "ComputerResource",
        "method": "move_mouse",
        "class_name": "Computer",
        "span_type": "TOOL",
    },
    {
        "package": "kernel.resources.browsers.computer",
        "object": "ComputerResource",
        "method": "press_key",
        "class_name": "Computer",
        "span_type": "TOOL",
    },
    {
        "package": "kernel.resources.browsers.computer",
        "object": "ComputerResource",
        "method": "scroll",
        "class_name": "Computer",
        "span_type": "TOOL",
    },
    {
        "package": "kernel.resources.browsers.computer",
        "object": "ComputerResource",
        "method": "type_text",
        "class_name": "Computer",
        "span_type": "TOOL",
    },
    {
        "package": "kernel.resources.browsers.playwright",
        "object": "PlaywrightResource",
        "method": "execute",
        "class_name": "Playwright",
    },
    {
        "package": "kernel.resources.browsers.process",
        "object": "ProcessResource",
        "method": "exec",
        "class_name": "Process",
        "span_type": "TOOL",
        "output_formatter": process_tool_output_formatter,
    },
    {
        "package": "kernel.resources.browsers.process",
        "object": "ProcessResource",
        "method": "kill",
        "class_name": "Process",
        "span_type": "TOOL",
        "output_formatter": process_tool_output_formatter,
    },
    {
        "package": "kernel.resources.browsers.process",
        "object": "ProcessResource",
        "method": "spawn",
        "class_name": "Process",
        "span_type": "TOOL",
        "output_formatter": process_tool_output_formatter,
    },
    {
        "package": "kernel.resources.browsers.process",
        "object": "ProcessResource",
        "method": "status",
        "class_name": "Process",
        "span_type": "TOOL",
        "output_formatter": process_tool_output_formatter,
    },
    {
        "package": "kernel.resources.browsers.process",
        "object": "ProcessResource",
        "method": "stdin",
        "class_name": "Process",
        "span_type": "TOOL",
        "output_formatter": process_tool_output_formatter,
    },
    {
        "package": "kernel.resources.browsers.process",
        "object": "ProcessResource",
        "method": "stdout_stream",
        "class_name": "Process",
        "span_type": "TOOL",
        "output_formatter": process_tool_output_formatter,
    },
]


def _with_wrapper(func):
    """Helper for providing tracer for wrapper functions. Includes metric collectors."""

    def wrapper(
        to_wrap,
    ):
        def wrapper(wrapped, instance, args, kwargs):
            return func(
                to_wrap,
                wrapped,
                instance,
                args,
                kwargs,
            )

        return wrapper

    return wrapper


@_with_wrapper
def _wrap(
    to_wrap,
    wrapped,
    instance,
    args,
    kwargs,
):
    with Laminar.start_as_current_span(
        f"{to_wrap.get('class_name')}.{to_wrap.get('method')}",
        span_type=to_wrap.get("span_type", "DEFAULT"),
    ) as span:
        input_kv = get_input_from_func_args(wrapped, True, args, kwargs)
        if "id" in input_kv:
            input_kv["session_id"] = input_kv.get("id")
            input_kv.pop("id")
        span.set_attribute(
            "lmnr.span.input",
            json_dumps(input_kv),
        )
        try:
            result = wrapped(*args, **kwargs)
        except Exception as e:  # pylint: disable=broad-except
            span.set_status(Status(StatusCode.ERROR))
            span.record_exception(e)
            raise
        output_formatter = to_wrap.get("output_formatter") or (lambda x: json_dumps(x))
        span.set_attribute("lmnr.span.output", output_formatter(result))
        return result


@_with_wrapper
async def _wrap_async(
    to_wrap,
    wrapped,
    instance,
    args,
    kwargs,
):
    with Laminar.start_as_current_span(
        f"{to_wrap.get('class_name')}.{to_wrap.get('method')}",
        span_type=to_wrap.get("span_type", "DEFAULT"),
    ) as span:
        input_kv = get_input_from_func_args(wrapped, True, args, kwargs)
        if "id" in input_kv:
            input_kv["session_id"] = input_kv.get("id")
            input_kv.pop("id")
        span.set_attribute(
            "lmnr.span.input",
            json_dumps(input_kv),
        )
        try:
            result = await wrapped(*args, **kwargs)
        except Exception as e:  # pylint: disable=broad-except
            span.set_status(Status(StatusCode.ERROR))
            span.record_exception(e)
            raise
        output_formatter = to_wrap.get("output_formatter") or (lambda x: json_dumps(x))
        span.set_attribute("lmnr.span.output", output_formatter(result))
        return result


@_with_wrapper
def _wrap_app_action(
    to_wrap,
    wrapped,
    instance,
    args,
    kwargs,
):
    """
    Wraps app.action() decorator factory to add tracing to action handlers.

    wrapped: the original `action` method
    args: (name,) - the action name
    kwargs: potentially {'name': ...}

    Returns a decorator that wraps handlers with tracing before registering them.
    """

    # Call the original action method to get the decorator
    original_decorator = wrapped(*args, **kwargs)

    # Get the action name from args
    action_name = args[0] if args else kwargs.get("name", "unknown")

    # Create a wrapper for the decorator that intercepts the handler
    def tracing_decorator(handler):
        # Apply the observe decorator to add tracing
        observed_handler = observe(
            name=f"action.{action_name}",
            span_type="DEFAULT",
        )(handler)

        # Create an additional wrapper to add post-execution logic
        if is_async(handler):

            @functools.wraps(handler)
            async def async_wrapper_with_flush(*handler_args, **handler_kwargs):
                # Execute the observed handler (tracing happens here)
                result = await observed_handler(*handler_args, **handler_kwargs)

                Laminar.flush()

                return result

            # Register the wrapper with the original decorator
            return original_decorator(async_wrapper_with_flush)
        else:

            @functools.wraps(handler)
            def sync_wrapper_with_flush(*handler_args, **handler_kwargs):
                # Execute the observed handler (tracing happens here)
                result = observed_handler(*handler_args, **handler_kwargs)

                Laminar.flush()

                return result

            # Register the wrapper with the original decorator
            return original_decorator(sync_wrapper_with_flush)

    return tracing_decorator


class KernelInstrumentor(BaseInstrumentor):
    def __init__(self):
        super().__init__()

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        for wrapped_method in WRAPPED_METHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            wrap_method = wrapped_method.get("method")

            try:
                wrap_function_wrapper(
                    wrap_package,
                    f"{wrap_object}.{wrap_method}",
                    _wrap(wrapped_method),
                )
            except (ModuleNotFoundError, AttributeError):
                pass  # that's ok, we don't want to fail if some methods do not exist

        for wrapped_method in WRAPPED_METHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = f"Async{wrapped_method.get('object')}"
            wrap_method = wrapped_method.get("method")
            try:
                wrap_function_wrapper(
                    wrap_package,
                    f"{wrap_object}.{wrap_method}",
                    _wrap_async(wrapped_method),
                )
            except (ModuleNotFoundError, AttributeError):
                pass  # that's ok, we don't want to fail if some methods do not exist

        try:
            wrap_function_wrapper(
                "kernel.app_framework",
                "KernelApp.action",
                _wrap_app_action({}),
            )
        except (ModuleNotFoundError, AttributeError):
            pass

    def _uninstrument(self, **kwargs):
        for wrapped_method in WRAPPED_METHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            try:
                unwrap(
                    f"{wrap_package}.{wrap_object}",
                    wrapped_method.get("method"),
                )
            except (ModuleNotFoundError, AttributeError):
                pass  # that's ok, we don't want to fail if some methods do not exist

        for wrapped_method in WRAPPED_METHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = f"Async{wrapped_method.get('object')}"
            try:
                unwrap(
                    f"{wrap_package}.{wrap_object}",
                    wrapped_method.get("method"),
                )
            except (ModuleNotFoundError, AttributeError):
                pass  # that's ok, we don't want to fail if some methods do not exist

        try:
            unwrap("kernel.app_framework.KernelApp", "action")
        except (ModuleNotFoundError, AttributeError):
            pass



import logging
import os
import traceback
from importlib.metadata import version

from opentelemetry import context as context_api
from .config import Config
from opentelemetry.semconv_ai import SpanAttributes

GEN_AI_SYSTEM = "gen_ai.system"
GEN_AI_SYSTEM_GROQ = "groq"

_PYDANTIC_VERSION = version("pydantic")

LMNR_TRACE_CONTENT = "LMNR_TRACE_CONTENT"


def set_span_attribute(span, name, value):
    if value is not None and value != "":
        span.set_attribute(name, value)


def should_send_prompts():
    return (
        os.getenv(LMNR_TRACE_CONTENT) or "true"
    ).lower() == "true" or context_api.get_value("override_enable_content_tracing")


def dont_throw(func):
    """
    A decorator that wraps the passed in function and logs exceptions instead of throwing them.

    @param func: The function to wrap
    @return: The wrapper function
    """
    # Obtain a logger specific to the function's module
    logger = logging.getLogger(func.__module__)

    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            logger.debug(
                "OpenLLMetry failed to trace in %s, error: %s",
                func.__name__,
                traceback.format_exc(),
            )
            if Config.exception_logger:
                Config.exception_logger(e)

    return wrapper


@dont_throw
def shared_metrics_attributes(response):
    response_dict = model_as_dict(response)

    common_attributes = Config.get_common_metrics_attributes()

    return {
        **common_attributes,
        GEN_AI_SYSTEM: GEN_AI_SYSTEM_GROQ,
        SpanAttributes.LLM_RESPONSE_MODEL: response_dict.get("model"),
    }


@dont_throw
def error_metrics_attributes(exception):
    return {
        GEN_AI_SYSTEM: GEN_AI_SYSTEM_GROQ,
        "error.type": exception.__class__.__name__,
    }


def model_as_dict(model):
    if _PYDANTIC_VERSION < "2.0.0":
        return model.dict()
    if hasattr(model, "model_dump"):
        return model.model_dump()
    elif hasattr(model, "parse"):  # Raw API response
        return model_as_dict(model.parse())
    else:
        return model


def should_emit_events() -> bool:
    """
    Checks if the instrumentation isn't using the legacy attributes
    and if the event logger is not None.
    """

    return not Config.use_legacy_attributes



import grpc
import re
import threading
from typing import Sequence
from urllib.parse import urlparse, urlunparse
from opentelemetry.sdk.trace.export import SpanExporter, SpanExportResult
from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.sdk._logs import ReadableLogRecord
from opentelemetry.sdk._logs.export import LogRecordExporter, LogRecordExportResult
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import (
    OTLPSpanExporter,
)
from opentelemetry.exporter.otlp.proto.grpc._log_exporter import (
    OTLPLogExporter,
)
from opentelemetry.exporter.otlp.proto.http import Compression as HTTPCompression
from opentelemetry.exporter.otlp.proto.http.trace_exporter import (
    OTLPSpanExporter as HTTPOTLPSpanExporter,
)
from opentelemetry.exporter.otlp.proto.http._log_exporter import (
    OTLPLogExporter as HTTPOTLPLogExporter,
)

from lmnr.sdk.log import get_default_logger
from lmnr.sdk.utils import from_env, get_otel_env_var, parse_otel_headers

logger = get_default_logger(__name__)


def _configure_exporter(
    base_url: str | None,
    port: int | None,
    api_key: str | None,
    timeout_seconds: int,
    force_http: bool,
) -> dict:
    """Configure common exporter settings for both span and log exporters.
    
    Args:
        base_url: Base URL for the exporter
        port: Port number for the exporter
        api_key: API key for authentication
        timeout_seconds: Timeout in seconds
        force_http: Whether to force HTTP protocol
        
    Returns:
        Dictionary with endpoint, headers, timeout, and force_http settings
    """
    url = base_url or from_env("LMNR_BASE_URL") or "https://api.lmnr.ai"
    url = url.rstrip("/")
    if match := re.search(r":(\d{1,5})$", url):
        url = url[: -len(match.group(0))]
        if port is None:
            port = int(match.group(1))
    if port is None:
        port = 443 if force_http else 8443
    final_url = f"{url}:{port or 443}"
    api_key = api_key or from_env("LMNR_PROJECT_API_KEY")
    endpoint = final_url
    timeout = timeout_seconds
    force_http_result = force_http
    
    if api_key:
        headers = (
            {"Authorization": f"Bearer {api_key}"}
            if force_http
            else {"authorization": f"Bearer {api_key}"}
        )
    else:
        if env_headers := get_otel_env_var("HEADERS"):
            headers = parse_otel_headers(env_headers)
        else:
            headers = {}
        if env_endpoint := get_otel_env_var("ENDPOINT"):
            if not base_url:
                endpoint = env_endpoint
                protocol = get_otel_env_var("PROTOCOL") or "grpc/protobuf"
                exporter_type = from_env("OTEL_EXPORTER") or "otlp_grpc"
                force_http_result = (
                    protocol in ("http/protobuf", "http/json")
                    or exporter_type == "otlp_http"
                )
            else:
                logger.warning(
                    "OTEL_ENDPOINT is set, but Laminar base URL is also set. Ignoring OTEL_ENDPOINT."
                )
    
    if not endpoint:
        raise ValueError(
            "Laminar base URL is not set and OTEL_ENDPOINT is not set. Please either\n"
            "- set the LMNR_BASE_URL environment variable\n"
            "- set the OTEL_ENDPOINT environment variable\n"
            "- pass the base_url parameter to Laminar.initialize"
        )
    
    return {
        "endpoint": endpoint,
        "headers": headers,
        "timeout": timeout,
        "force_http": force_http_result,
    }


def _normalize_http_endpoint(endpoint: str, default_path: str) -> str:
        """
        Normalize HTTP endpoint URL by adding the given default path if no path is present.

        Args:
            endpoint: The endpoint URL to normalize

        Returns:
            The normalized endpoint URL with the given default path if needed
        """
        try:
            parsed = urlparse(endpoint)
            # Check if there's no path or only a trailing slash
            if not parsed.path or parsed.path == "/":
                # Add the given path to the endpoint
                new_parsed = parsed._replace(path=default_path)
                normalized_url = urlunparse(new_parsed)
                logger.info(
                    f"No path found in HTTP endpoint URL. "
                    f"Adding default path {default_path}: {endpoint} -> {normalized_url}"
                )
                return normalized_url
            return endpoint
        except Exception as e:
            logger.warning(
                f"Failed to parse endpoint URL '{endpoint}': {e}. Using as-is."
            )
            return endpoint


class LaminarSpanExporter(SpanExporter):
    instance: OTLPSpanExporter | HTTPOTLPSpanExporter
    endpoint: str
    headers: dict[str, str]
    timeout: float
    force_http: bool
    _instance_lock: threading.RLock

    def __init__(
        self,
        base_url: str | None = None,
        port: int | None = None,
        api_key: str | None = None,
        timeout_seconds: int = 30,
        force_http: bool = False,
    ):
        self._instance_lock = threading.RLock()
        config = _configure_exporter(base_url, port, api_key, timeout_seconds, force_http)
        self.endpoint = config["endpoint"]
        self.headers = config["headers"]
        self.timeout = config["timeout"]
        self.force_http = config["force_http"]
        self._init_instance()

    def _init_instance(self):
        # Create new instance first (outside critical section for performance)
        if self.force_http:
            # Normalize HTTP endpoint to ensure it has a path
            http_endpoint = _normalize_http_endpoint(self.endpoint, "/v1/traces")
            new_instance = HTTPOTLPSpanExporter(
                endpoint=http_endpoint,
                headers=self.headers,
                compression=HTTPCompression.Gzip,
                timeout=self.timeout,
            )
        else:
            new_instance = OTLPSpanExporter(
                endpoint=self.endpoint,
                headers=self.headers,
                timeout=self.timeout,
                compression=grpc.Compression.Gzip,
            )

        with self._instance_lock:
            old_instance: OTLPSpanExporter | HTTPOTLPSpanExporter | None = getattr(
                self, "instance", None
            )
            if old_instance is not None:
                try:
                    old_instance.shutdown()
                except Exception as e:
                    logger.warning(f"Error shutting down old exporter instance: {e}")
            self.instance = new_instance

    def export(self, spans: list[ReadableSpan]) -> SpanExportResult:
        with self._instance_lock:
            return self.instance.export(spans)

    def shutdown(self) -> None:
        with self._instance_lock:
            return self.instance.shutdown()

    def force_flush(self, timeout_millis: int = 30000) -> bool:
        with self._instance_lock:
            return self.instance.force_flush(timeout_millis)

class LaminarLogExporter(LogRecordExporter):
    """Log exporter for Laminar that sends logs to the OTLP endpoint."""

    instance: OTLPLogExporter | HTTPOTLPLogExporter
    endpoint: str
    headers: dict[str, str]
    timeout: float
    force_http: bool

    def __init__(
        self,
        base_url: str | None = None,
        port: int | None = None,
        api_key: str | None = None,
        timeout_seconds: int = 30,
        force_http: bool = False,
    ):
        config = _configure_exporter(base_url, port, api_key, timeout_seconds, force_http)
        self.endpoint = config["endpoint"]
        self.headers = config["headers"]
        self.timeout = config["timeout"]
        self.force_http = config["force_http"]
        self._init_instance()

    def _init_instance(self):
        # Create new instance first (outside critical section for performance)
        if self.force_http:
            # Normalize HTTP endpoint to ensure it has a path
            http_endpoint = _normalize_http_endpoint(self.endpoint, "/v1/logs")
            new_instance = HTTPOTLPLogExporter(
                endpoint=http_endpoint,
                headers=self.headers,
                compression=HTTPCompression.Gzip,
                timeout=self.timeout,
            )
        else:
            new_instance = OTLPLogExporter(
                endpoint=self.endpoint,
                headers=self.headers,
                timeout=self.timeout,
                compression=grpc.Compression.Gzip,
            )

        # Atomic swap with proper cleanup
        old_instance: OTLPLogExporter | HTTPOTLPLogExporter | None = getattr(
            self, "instance", None
        )
        if old_instance is not None:
            try:
                old_instance.shutdown()
            except Exception as e:
                logger.warning(f"Error shutting down old exporter instance: {e}")
        self.instance = new_instance

    def export(self, batch: Sequence[ReadableLogRecord]) -> LogRecordExportResult:
        return self.instance.export(batch)

    def shutdown(self) -> None:
        return self.instance.shutdown()

    def force_flush(self, timeout_millis: int = 30000) -> bool:
        return self.instance.force_flush(timeout_millis)



from lmnr.opentelemetry_lib.tracing.context import get_event_attributes_from_context
from ..shared import _set_span_attribute
from ..shared.event_emitter import emit_event
from ..shared.event_models import ChoiceEvent
from ..utils import should_emit_events
from opentelemetry.semconv.attributes.error_attributes import ERROR_TYPE
from opentelemetry.semconv_ai import SpanAttributes
from opentelemetry.trace import Status, StatusCode
from typing_extensions import override

from openai import AssistantEventHandler


class EventHandlerWrapper(AssistantEventHandler):
    _current_text_index = 0
    _prompt_tokens = 0
    _completion_tokens = 0

    def __init__(self, original_handler, span):
        super().__init__()
        self._original_handler = original_handler
        self._span = span

    @override
    def on_end(self):
        _set_span_attribute(
            self._span,
            SpanAttributes.LLM_USAGE_PROMPT_TOKENS,
            self._prompt_tokens,
        )
        _set_span_attribute(
            self._span,
            SpanAttributes.LLM_USAGE_COMPLETION_TOKENS,
            self._completion_tokens,
        )
        self._original_handler.on_end()
        self._span.end()

    @override
    def on_event(self, event):
        self._original_handler.on_event(event)

    @override
    def on_run_step_created(self, run_step):
        self._original_handler.on_run_step_created(run_step)

    @override
    def on_run_step_delta(self, delta, snapshot):
        self._original_handler.on_run_step_delta(delta, snapshot)

    @override
    def on_run_step_done(self, run_step):
        if run_step.usage:
            self._prompt_tokens += run_step.usage.prompt_tokens
            self._completion_tokens += run_step.usage.completion_tokens
        self._original_handler.on_run_step_done(run_step)

    @override
    def on_tool_call_created(self, tool_call):
        self._original_handler.on_tool_call_created(tool_call)

    @override
    def on_tool_call_delta(self, delta, snapshot):
        self._original_handler.on_tool_call_delta(delta, snapshot)

    @override
    def on_tool_call_done(self, tool_call):
        self._original_handler.on_tool_call_done(tool_call)

    @override
    def on_exception(self, exception: Exception):
        self._span.set_attribute(ERROR_TYPE, exception.__class__.__name__)
        self._span.record_exception(
            exception, attributes=get_event_attributes_from_context()
        )
        self._span.set_status(Status(StatusCode.ERROR, str(exception)))
        self._original_handler.on_exception(exception)

    @override
    def on_timeout(self):
        self._original_handler.on_timeout()

    @override
    def on_message_created(self, message):
        self._original_handler.on_message_created(message)

    @override
    def on_message_delta(self, delta, snapshot):
        self._original_handler.on_message_delta(delta, snapshot)

    @override
    def on_message_done(self, message):
        _set_span_attribute(
            self._span,
            f"gen_ai.response.{self._current_text_index}.id",
            message.id,
        )
        emit_event(
            ChoiceEvent(
                index=self._current_text_index,
                message={
                    "content": [item.model_dump() for item in message.content],
                    "role": message.role,
                },
            )
        )
        self._original_handler.on_message_done(message)
        self._current_text_index += 1

    @override
    def on_text_created(self, text):
        self._original_handler.on_text_created(text)

    @override
    def on_text_delta(self, delta, snapshot):
        self._original_handler.on_text_delta(delta, snapshot)

    @override
    def on_text_done(self, text):
        self._original_handler.on_text_done(text)
        if not should_emit_events():
            _set_span_attribute(
                self._span,
                f"{SpanAttributes.LLM_COMPLETIONS}.{self._current_text_index}.role",
                "assistant",
            )
            _set_span_attribute(
                self._span,
                f"{SpanAttributes.LLM_COMPLETIONS}.{self._current_text_index}.content",
                text.value,
            )

    @override
    def on_image_file_done(self, image_file):
        self._original_handler.on_image_file_done(image_file)



from argparse import ArgumentParser, _SubParsersAction
import asyncio

from lmnr.cli.datasets import handle_datasets_command
from lmnr.cli.dev import run_dev
from lmnr.cli.discover import run_discover
from lmnr.cli.evals import run_evaluation
from lmnr.cli.rules import add_cursor_rules
from lmnr.sdk.log import get_default_logger
from lmnr.sdk.utils import from_env

LOG = get_default_logger(__name__)
EVAL_DIR = "evals"
DEFAULT_DATASET_PULL_BATCH_SIZE = 100
DEFAULT_DATASET_PUSH_BATCH_SIZE = 100


def setup_eval_parser(subparsers: _SubParsersAction) -> None:
    """Setup the eval subcommand parser."""
    parser_eval = subparsers.add_parser(
        "eval",
        description="Run an evaluation",
        help="Run an evaluation",
    )
    parser_eval.add_argument(
        "file",
        nargs="*",
        help="Files or a file containing the evaluation to run. "
        + "If no file name is provided, all evaluation files in the `evals` directory are run as long "
        + "as they match *_eval.py or eval_*.py",
        default=[],
    )
    parser_eval.add_argument(
        "--continue-on-error",
        action="store_true",
        default=False,
        help="Continue execution upon errors",
    )
    parser_eval.add_argument(
        "--output-file",
        help="Output file to write the results to. Outputs are written in JSON format.",
        nargs="?",
    )
    parser_eval.add_argument(
        "--frontend-port",
        help="[Optional] Port for the frontend when running locally. "
        + "If no port is provided, defaults to '5667'.",
        type=int,
        default=None,
    )


def setup_dev_parser(subparsers: _SubParsersAction) -> None:
    """Setup the dev subcommand parser."""
    parser_dev = subparsers.add_parser(
        "dev",
        description="Start a rollout development session for interactive LLM debugging",
        help="Start a rollout development session",
    )
    parser_dev.add_argument(
        "file",
        help="Path to Python file containing rollout entrypoint function(s)",
    )
    parser_dev.add_argument(
        "--function",
        "-f",
        help="[Optional] Specific function name to use as entrypoint. "
        + "If not provided, automatically discovers the rollout entrypoint. "
        + "Required if multiple entrypoints exist in the file.",
        default=None,
    )
    parser_dev.add_argument(
        "--grpc-port",
        help="[Optional] Port to use for the gRPC server. "
        + "If no port is provided, the port defaults to '8443'.",
        type=int,
        default=8443,
    )
    parser_dev.add_argument(
        "--frontend-port",
        help="[Optional] Port for the frontend when running locally. "
        + "If no port is provided, defaults to '5667'.",
        type=int,
        default=None,
    )
    setup_laminar_args(parser_dev)


def setup_discover_parser(subparsers: _SubParsersAction) -> None:
    """Setup the discover subcommand parser."""
    parser_discover = subparsers.add_parser(
        "discover",
        description="Internal command for discovering rollout function metadata",
        help="Discover rollout function metadata (internal command for CLI)",
    )

    # Mutually exclusive group for file vs module
    mode_group = parser_discover.add_mutually_exclusive_group(required=True)
    mode_group.add_argument(
        "--file",
        help="Path to Python file containing rollout function(s)",
    )
    mode_group.add_argument(
        "--module",
        "-m",
        help="Python module path (e.g., src.myfile)",
    )

    parser_discover.add_argument(
        "--function",
        "-f",
        help="Specific function name to discover. If not provided, discovers all entrypoints.",
        default=None,
    )


def setup_add_cursor_rules_parser(subparsers: _SubParsersAction) -> None:
    """Setup the add-cursor-rules subcommand parser."""
    subparsers.add_parser(
        "add-cursor-rules",
        description="Download laminar.mdc file and add it to .cursor/rules",
        help="Download laminar.mdc file and add it to .cursor/rules",
    )


def setup_laminar_args(parser: ArgumentParser) -> None:
    """Setup the laminar arguments parser."""
    parser.add_argument(
        "--project-api-key",
        help="[Optional] Project API key to use for the command. "
        + "If no project API key is provided, the project API key will be read "
        + "from the environment variable LMNR_PROJECT_API_KEY.",
        default=from_env("LMNR_PROJECT_API_KEY"),
    )
    parser.add_argument(
        "--base-url",
        help="[Optional] Base URL to use for the command. "
        + "If no base URL is provided, the base URL will be read from the "
        + "'LMNR_BASE_URL' environment variable or we default to 'https://api.lmnr.ai'.",
        default=from_env("LMNR_BASE_URL") or "https://api.lmnr.ai",
    )
    parser.add_argument(
        "--port",
        help="[Optional] Port to use for the command. "
        + "If no port is provided, the port defaults to '443'.",
        type=int,
    )


def setup_datasets_list_parser(subparsers: _SubParsersAction) -> None:
    """Setup the datasets list subcommand parser."""
    subparsers.add_parser(
        "list",
        description="List datasets",
        help="List datasets",
    )


def setup_datasets_push_parser(subparsers: _SubParsersAction) -> None:
    """Setup the datasets push subcommand parser."""
    parser_datasets_push: ArgumentParser = subparsers.add_parser(
        "push",
        description="Push datapoints to an existing dataset",
        help="Push datapoints to an existing dataset",
    )
    parser_datasets_push.add_argument(
        "--name",
        "-n",
        help="Name of the dataset to push data to. Exactly one of name or id must be provided.",
        default=None,
    )
    parser_datasets_push.add_argument(
        "--id",
        help="ID of the dataset to push data to. Exactly one of name or id must be provided.",
        default=None,
    )
    parser_datasets_push.add_argument(
        "paths",
        nargs="*",
        help="Paths to the files or directories containing the data to push to the dataset. "
        + "Supported formats: JSON, CSV, JSONL",
    )
    parser_datasets_push.add_argument(
        "-r",
        "--recursive",
        action="store_true",
        default=False,
        help="Recursively read all files in the directories and their subdirectories.",
    )
    parser_datasets_push.add_argument(
        "--batch-size",
        type=int,
        help="Batch size to push data in. If no batch size is provided, "
        + f"data is pushed in batches of '{DEFAULT_DATASET_PUSH_BATCH_SIZE}'.",
        default=DEFAULT_DATASET_PUSH_BATCH_SIZE,
    )


def setup_datasets_pull_parser(subparsers: _SubParsersAction) -> None:
    """Setup the datasets pull subcommand parser."""
    parser_datasets_pull: ArgumentParser = subparsers.add_parser(
        "pull",
        description="Pull data from a dataset",
        help="Pull data from a dataset",
    )
    parser_datasets_pull.add_argument(
        "--name",
        "-n",
        help="Name of the dataset to pull data from",
        default=None,
    )
    parser_datasets_pull.add_argument(
        "--id",
        help="ID of the dataset to pull data from",
        default=None,
    )
    parser_datasets_pull.add_argument(
        "output_path",
        help="Path to the file to save the data to. "
        + "If no path is provided, data is printed to the console in the format "
        + "specified by '--output-format'.",
        nargs="?",
    )
    parser_datasets_pull.add_argument(
        "--output-format",
        choices=["json", "csv", "jsonl"],
        help="Output format to save the data to. "
        + "If no format is provided, it is inferred from the file extension.",
    )
    parser_datasets_pull.add_argument(
        "--batch-size",
        type=int,
        help="Batch size to pull data in. If no batch size is provided, "
        + f"data is pulled in batches of '{DEFAULT_DATASET_PULL_BATCH_SIZE}'.",
        default=DEFAULT_DATASET_PULL_BATCH_SIZE,
    )
    parser_datasets_pull.add_argument(
        "--limit",
        type=int,
        help="Limit the number of data points to pull. "
        + "If no limit is provided, all data points are pulled.",
    )
    parser_datasets_pull.add_argument(
        "--offset",
        type=int,
        help="Offset the number of data points to pull. "
        + "If no offset is provided, data is pulled from the beginning.",
    )


def setup_datasets_create_parser(subparsers: _SubParsersAction) -> None:
    """Setup the datasets create subcommand parser."""
    parser_datasets_create: ArgumentParser = subparsers.add_parser(
        "create",
        description="Create a dataset from input files and download it in Laminar format",
        help="Create a dataset from input files and download it in Laminar format",
    )
    parser_datasets_create.add_argument(
        "name",
        help="Name of the dataset to create",
    )
    parser_datasets_create.add_argument(
        "paths",
        nargs="+",
        help="Paths to the files or directories containing the data to push to the dataset. "
        + "Supported formats: JSON, CSV, JSONL",
    )
    parser_datasets_create.add_argument(
        "-o",
        "--output-file",
        required=True,
        help="Path to the file to save the pulled data to",
    )
    parser_datasets_create.add_argument(
        "--output-format",
        choices=["json", "csv", "jsonl"],
        help="Output format to save the data to. "
        + "If no format is provided, it is inferred from the output file extension.",
    )
    parser_datasets_create.add_argument(
        "-r",
        "--recursive",
        action="store_true",
        default=False,
        help="Recursively read all files in the directories and their subdirectories.",
    )
    parser_datasets_create.add_argument(
        "--batch-size",
        type=int,
        help="Batch size to push/pull data in. If no batch size is provided, "
        + f"data is processed in batches of '{DEFAULT_DATASET_PUSH_BATCH_SIZE}'.",
        default=DEFAULT_DATASET_PUSH_BATCH_SIZE,
    )


def setup_datasets_parser(subparsers: _SubParsersAction) -> None:
    """Setup the datasets subcommand parser and its subcommands."""
    parser_datasets: ArgumentParser = subparsers.add_parser(
        "datasets",
        description="Manage datasets",
        help="Manage datasets",
    )

    setup_laminar_args(parser_datasets)

    parser_datasets_subparsers = parser_datasets.add_subparsers(
        title="command",
        dest="command",
    )

    # Setup all dataset subcommands
    setup_datasets_list_parser(parser_datasets_subparsers)
    setup_datasets_push_parser(parser_datasets_subparsers)
    setup_datasets_pull_parser(parser_datasets_subparsers)
    setup_datasets_create_parser(parser_datasets_subparsers)


def cli() -> None:
    """Main CLI entry point."""
    parser = ArgumentParser(
        prog="lmnr",
        description="CLI for Laminar. "
        + "Call `lmnr [subcommand] --help` for more information on each subcommand.",
    )

    subparsers = parser.add_subparsers(title="subcommands", dest="subcommand")

    # Setup all subcommand parsers
    setup_eval_parser(subparsers)
    setup_dev_parser(subparsers)
    setup_discover_parser(subparsers)
    setup_add_cursor_rules_parser(subparsers)
    setup_datasets_parser(subparsers)

    # Parse arguments and dispatch to appropriate handler
    parsed = parser.parse_args()

    if parsed.subcommand == "eval":
        asyncio.run(run_evaluation(parsed))
    elif parsed.subcommand == "dev":
        asyncio.run(run_dev(parsed))
    elif parsed.subcommand == "discover":
        run_discover(parsed)
    elif parsed.subcommand == "add-cursor-rules":
        add_cursor_rules()
    elif parsed.subcommand == "datasets":
        asyncio.run(handle_datasets_command(parsed))
    else:
        parser.print_help()



import logging
import time
from collections.abc import Iterable

from opentelemetry import context as context_api

from lmnr.opentelemetry_lib.tracing.context import get_event_attributes_from_context
from ..shared import (
    OPENAI_LLM_USAGE_TOKEN_TYPES,
    _get_openai_base_url,
    _set_client_attributes,
    _set_request_attributes,
    _set_response_attributes,
    _set_span_attribute,
    _token_type,
    metric_shared_attributes,
    model_as_dict,
    propagate_trace_context,
)
from ..shared.config import Config
from ..shared.event_emitter import emit_event
from ..shared.event_models import (
    ChoiceEvent,
    MessageEvent,
)
from ..utils import (
    _with_embeddings_telemetry_wrapper,
    dont_throw,
    is_openai_v1,
    should_emit_events,
    should_send_prompts,
    start_as_current_span_async,
)
from opentelemetry.instrumentation.utils import _SUPPRESS_INSTRUMENTATION_KEY
from opentelemetry.metrics import Counter, Histogram
from opentelemetry.semconv.attributes.error_attributes import ERROR_TYPE
from opentelemetry.semconv_ai import (
    SUPPRESS_LANGUAGE_MODEL_INSTRUMENTATION_KEY,
    LLMRequestTypeValues,
    SpanAttributes,
)
from opentelemetry.trace import SpanKind, Status, StatusCode

from openai._legacy_response import LegacyAPIResponse
from openai.types.create_embedding_response import CreateEmbeddingResponse

SPAN_NAME = "openai.embeddings"
LLM_REQUEST_TYPE = LLMRequestTypeValues.EMBEDDING

logger = logging.getLogger(__name__)


@_with_embeddings_telemetry_wrapper
def embeddings_wrapper(
    tracer,
    token_counter: Counter,
    vector_size_counter: Counter,
    duration_histogram: Histogram,
    exception_counter: Counter,
    wrapped,
    instance,
    args,
    kwargs,
):
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY) or context_api.get_value(
        SUPPRESS_LANGUAGE_MODEL_INSTRUMENTATION_KEY
    ):
        return wrapped(*args, **kwargs)

    with tracer.start_as_current_span(
        name=SPAN_NAME,
        kind=SpanKind.CLIENT,
        attributes={SpanAttributes.LLM_REQUEST_TYPE: LLM_REQUEST_TYPE.value},
    ) as span:
        _handle_request(span, kwargs, instance)

        try:
            # record time for duration
            start_time = time.time()
            response = wrapped(*args, **kwargs)
            end_time = time.time()
        except Exception as e:  # pylint: disable=broad-except
            end_time = time.time()
            duration = end_time - start_time if "start_time" in locals() else 0
            attributes = {
                "error.type": e.__class__.__name__,
            }

            # if there are legal duration, record it
            if duration > 0 and duration_histogram:
                duration_histogram.record(duration, attributes=attributes)
            if exception_counter:
                exception_counter.add(1, attributes=attributes)

            span.set_attribute(ERROR_TYPE, e.__class__.__name__)
            attributes = get_event_attributes_from_context()
            span.record_exception(e, attributes=attributes)
            span.set_status(Status(StatusCode.ERROR, str(e)))
            span.end()

            raise

        duration = end_time - start_time

        _handle_response(
            response,
            span,
            instance,
            token_counter,
            vector_size_counter,
            duration_histogram,
            duration,
        )

        return response


@_with_embeddings_telemetry_wrapper
async def aembeddings_wrapper(
    tracer,
    token_counter: Counter,
    vector_size_counter: Counter,
    duration_histogram: Histogram,
    exception_counter: Counter,
    wrapped,
    instance,
    args,
    kwargs,
):
    if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY) or context_api.get_value(
        SUPPRESS_LANGUAGE_MODEL_INSTRUMENTATION_KEY
    ):
        return await wrapped(*args, **kwargs)

    async with start_as_current_span_async(
        tracer=tracer,
        name=SPAN_NAME,
        kind=SpanKind.CLIENT,
        attributes={SpanAttributes.LLM_REQUEST_TYPE: LLM_REQUEST_TYPE.value},
    ) as span:
        _handle_request(span, kwargs, instance)

        try:
            # record time for duration
            start_time = time.time()
            response = await wrapped(*args, **kwargs)
            end_time = time.time()
        except Exception as e:  # pylint: disable=broad-except
            end_time = time.time()
            duration = end_time - start_time if "start_time" in locals() else 0
            attributes = {
                "error.type": e.__class__.__name__,
            }

            # if there are legal duration, record it
            if duration > 0 and duration_histogram:
                duration_histogram.record(duration, attributes=attributes)
            if exception_counter:
                exception_counter.add(1, attributes=attributes)

            span.set_attribute(ERROR_TYPE, e.__class__.__name__)
            attributes = get_event_attributes_from_context()
            span.record_exception(e, attributes=attributes)
            span.set_status(Status(StatusCode.ERROR, str(e)))
            span.end()

            raise

        duration = end_time - start_time

        _handle_response(
            response,
            span,
            instance,
            token_counter,
            vector_size_counter,
            duration_histogram,
            duration,
        )

        return response


@dont_throw
def _handle_request(span, kwargs, instance):
    _set_request_attributes(span, kwargs, instance)

    if should_emit_events():
        _emit_embeddings_message_event(kwargs.get("input"))
    else:
        if should_send_prompts():
            _set_prompts(span, kwargs.get("input"))

    _set_client_attributes(span, instance)

    if Config.enable_trace_context_propagation:
        propagate_trace_context(span, kwargs)


@dont_throw
def _handle_response(
    response,
    span,
    instance=None,
    token_counter=None,
    vector_size_counter=None,
    duration_histogram=None,
    duration=None,
):
    if is_openai_v1():
        response_dict = model_as_dict(response)
    else:
        response_dict = response
    # metrics record
    _set_embeddings_metrics(
        instance,
        token_counter,
        vector_size_counter,
        duration_histogram,
        response_dict,
        duration,
    )
    # span attributes
    _set_response_attributes(span, response_dict)

    # emit events
    if should_emit_events():
        _emit_embeddings_choice_event(response)


def _set_embeddings_metrics(
    instance,
    token_counter,
    vector_size_counter,
    duration_histogram,
    response_dict,
    duration,
):
    shared_attributes = metric_shared_attributes(
        response_model=response_dict.get("model") or None,
        operation="embeddings",
        server_address=_get_openai_base_url(instance),
    )

    # token count metrics
    usage = response_dict.get("usage")
    if usage and token_counter:
        for name, val in usage.items():
            if name in OPENAI_LLM_USAGE_TOKEN_TYPES:
                if val is None:
                    logging.error(f"Received None value for {name} in usage")
                    continue
                attributes_with_token_type = {
                    **shared_attributes,
                    SpanAttributes.LLM_TOKEN_TYPE: _token_type(name),
                }
                token_counter.record(val, attributes=attributes_with_token_type)

    # vec size metrics
    # should use counter for vector_size?
    vec_embedding = (response_dict.get("data") or [{}])[0].get("embedding", [])
    vec_size = len(vec_embedding)
    if vector_size_counter:
        vector_size_counter.add(vec_size, attributes=shared_attributes)

    # duration metrics
    if duration and isinstance(duration, (float, int)) and duration_histogram:
        duration_histogram.record(duration, attributes=shared_attributes)


def _set_prompts(span, prompt):
    if not span.is_recording() or not prompt:
        return

    if isinstance(prompt, list):
        for i, p in enumerate(prompt):
            _set_span_attribute(span, f"{SpanAttributes.LLM_PROMPTS}.{i}.content", p)
    else:
        _set_span_attribute(
            span,
            f"{SpanAttributes.LLM_PROMPTS}.0.content",
            prompt,
        )


def _emit_embeddings_message_event(embeddings) -> None:
    if isinstance(embeddings, str):
        emit_event(MessageEvent(content=embeddings))
    elif isinstance(embeddings, Iterable):
        for i in embeddings:
            emit_event(MessageEvent(content=i))


def _emit_embeddings_choice_event(response) -> None:
    if isinstance(response, CreateEmbeddingResponse):
        for embedding in response.data:
            emit_event(
                ChoiceEvent(
                    index=embedding.index,
                    message={"content": embedding.embedding, "role": "assistant"},
                )
            )

    elif isinstance(response, LegacyAPIResponse):
        parsed_response = response.parse()
        for embedding in parsed_response.data:
            emit_event(
                ChoiceEvent(
                    index=embedding.index,
                    message={"content": embedding.embedding, "role": "assistant"},
                )
            )



from copy import deepcopy
from typing import Any

import traceback

from opentelemetry.trace import Span
from opentelemetry.util.types import AttributeValue
from pydantic import BaseModel

from lmnr.sdk.log import get_default_logger

logger = get_default_logger(__name__)


def dont_throw(func):
    def wrapper(*args, **kwargs):
        logger = get_default_logger(func.__module__)
        try:
            return func(*args, **kwargs)
        except Exception:
            logger.debug(
                "Laminar failed to trace in %s, error: %s",
                func.__name__,
                traceback.format_exc(),
            )
            return None

    return wrapper


def set_span_attribute(
    span: Span, attribute_name: str, attribute_value: AttributeValue
):
    if attribute_value is not None and attribute_value != "":
        span.set_attribute(attribute_name, attribute_value)


def to_dict(obj: Any) -> dict:
    try:
        if isinstance(obj, BaseModel):
            return obj.model_dump()
        elif isinstance(obj, dict):
            return deepcopy(obj)
        elif obj is None:
            return {}
        else:
            return dict(obj)
    except Exception as e:
        logger.debug(f"Error converting to dict: {obj}, error: {e}")
        return {}


def extract_json_schema(schema: dict | BaseModel) -> dict:
    if isinstance(schema, dict):
        return schema
    elif hasattr(schema, "model_json_schema") and callable(schema.model_json_schema):
        return schema.model_json_schema()
    else:
        return {}
